import requests
import yaml
import os
import logging
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)


# inglobata in Askdata come metodo
class SignUp:

    def __init__(self, Askdata):

        self.username = Askdata.username
        self._token = Askdata._token
        self._env = Askdata._env
        self._domainlogin = Askdata._domainlogin.upper()

        self._headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        if self._env == 'dev':
            self.base_url_security = url_list['BASE_URL_SECURITY_DEV']
        if self._env == 'qa':
            self.base_url_security = url_list['BASE_URL_SECURITY_QA']
        if self._env == 'prod':
            self.base_url_security = url_list['BASE_URL_SECURITY_PROD']


    def signup_user(self, username: str, password: str, firstname='-', secondname='-', title='-') -> dict:

        data = {
            "username": username,
            "password": password,
            "firstName": firstname,
            "lastName": secondname,
            "mobile": "-",
            "title": title,
            "email": username
        }

        s = requests.Session()
        s.keep_alive = False
        #s.config['keep_alive'] = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = self.base_url_security + '/domain/' + self._domainlogin + '/usersignup'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        logging.info("--- ------------ ----")
        logging.info("---- add user ---> {}  --- PENDING".format(username))

        return r.json()


import requests
import yaml
import os
import pandas as pd
import logging
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from askdata.askdata_client import Agent
import numpy as np

_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)



class Catalog:

    _language = None
    _agentId = None
    _domain = None

    def __init__(self, env ,token):

        self._headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + token
        }

        if env == 'dev':
            self._base_url_cat = url_list['BASE_URL_FEED_DEV']
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_DEV']
        if env == 'qa':
            self._base_url_cat = url_list['BASE_URL_FEED_QA']
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_QA']
        if env == 'prod':
            self._base_url_cat = url_list['BASE_URL_FEED_PROD']
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_PROD']

    def create_catalog(self, name: str, icon="https://storage.googleapis.com/askdata/interactions/icoInteractionDiscovery.png" ):

        data = {"type": "CUSTOM",
                   "name": name,
                   "icon": icon,
                   "lang": self._language}

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartfeed/agents/' + self._agentId + '/discovery-entry'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()

        logging.info('create catalog: {} -----> with Id {}'.format(str(r.json()['title']), str(r.json()['id'])))

        return r.json()['id']

    def load_catalogs(self, empty=True) -> pd.DataFrame:
        """
        Get all bookmarks of the agent

        :param empty: empty = True is for including all bookmarks (catalog) also empty
        :return: DataFrame
        """
        if empty:
            flag = 'true'
        else:
            flag = 'false'

        authentication_url = self._base_url_askdata + '/smartfeed/' + self._domain + '/discovery?emptyIncluded=' + flag
        r = requests.get(url=authentication_url, headers=self._headers)
        r.raise_for_status()
        df_catalogs = pd.DataFrame(r.json()['discovery'])

        return df_catalogs

    def create_query(self, query:str, entryid:str, execute=False)-> str:

        data = {
            "type": "text",
            "payload": query,
            "title": query,
            "lang": self._language
            }

        if execute:
            flag_ex = 'true'
        else:
            flag_ex = 'false'

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartfeed/agents/' + self._agentId + '/discovery-entry/' + entryid + '/queries?execute=' + flag_ex
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()

        logging.info('create query: "{}" in catalog: {}'.format(str(r.json()['payload']), entryid))

        return r.json()['id']

    def get_query_from_catalog(self,entryid: str) -> list:

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartfeed/agents/' + self._agentId + '/discovery-entry/' + entryid + '/queries'
        r = s.get(url=authentication_url, headers=self._headers)
        r.raise_for_status()

        return r.json()

    def delete_catalog(self,entryid):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartfeed/agents/' + self._agentId + '/discovery-entry/' + entryid
        r = s.delete(url=authentication_url, headers=self._headers)
        r.raise_for_status()

        logging.info('deleted catalog: {}'.format(entryid))
        return r

    def delete_query(self, entryid: str, queryid: str):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartfeed/agents/' + self._agentId + '/discovery-entry/' + entryid + '/queries/' + queryid
        r = s.delete(url=authentication_url, headers=self._headers)
        r.raise_for_status()

        logging.info('deleted query: {} from catalog: {}'.format(queryid, entryid))
        return r

    @staticmethod
    def delete_all_queries_catalog(agent: 'Agent', entry_id: str):
        queries_list = agent.get_query_from_catalog(entry_id)
        for query in queries_list:
            agent.delete_query(entryid=entry_id, queryid=query['id'])
        logging.info('All queries deleted from catalog with entry_id : {}'.format(entry_id))



from askdata.smartgraph import smart_synonyms

if __name__ == "__main__":

    word = "home"
    lang = "en"
    initial_synonyms_list = []

    synonyms, business_name = smart_synonyms(word=word, lang=lang, initial_synonyms_list=initial_synonyms_list)

    print("Synonyms: ")
    print(synonyms)
    print()
    print("Business name:")
    print(business_name)



''' Salesforce API contacts to dataframe '''

import pandas as pd
import requests
import json
import time
import datetime
import os
import sys
import pytz
import numpy as np

from simple_salesforce import Salesforce

# import credentials from config file
import config

# import custom functions
import functions

# set timezone
os.environ['TZ'] = 'America/Los_Angeles'
time.tzset()

# set pandas display options
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)

# timestamp
start_time = time.time()

# instantiate Salesforce connection
sf = Salesforce(username=config.username, password=config.password, security_token=config.security_token)

# get all contacts
contacts = sf.query_all("SELECT Id, Name, Salutation, FirstName, LastName, Title, Company, Phone, MobilePhone, Email, MailingStreet, MailingCity, MailingState, MailingPostalCode, MailingCountry, OtherStreet, OtherCity, OtherState, OtherPostalCode, OtherCountry, Birthdate, Description, LeadSource, Status, Lead_Source_Detail__c, ConvertedAccountId, ConvertedAccountName, ConvertedContactId, ConvertedContactName, ConvertedDate, ConvertedOpportunityId, ConvertedOpportunityName, Do_Not_Call, DoNotCallReason, EmailBouncedDate, EmailBouncedReason, Fax, First_Responded__c, HasOptedOutOfEmail, HasOptedOutOfFax, IsEmailBounced, Jigsaw, JigsawContactId, LastCURequestDate, LastCUUpdateDate, LastModifiedById, LastModifiedDate, LastActivityDate, LastViewedDate, Lead_Source_Detail__c, Lead_Source__c, MailingLatitude, MailingLongitude, MasterRecordId, MobilePhone_verified__c, NumberOfEmployees, NumberofLocations__c, OtherLatitude, OtherLongitude, OwnerId, Phone_verified__c, Salutation_c__c, SystemModstamp, Title_c__c, Type, Website, CreatedDate, CreatedById, LastModifiedDate, LastModifiedById, SystemModstamp, IsDeleted FROM Contact")

# convert to dataframe
contacts_df = pd.DataFrame(contacts['records']).drop(columns='attributes')

# add additional columns
contacts_df['CreatedDate'] = pd.to_datetime(contacts_df['CreatedDate'])
contacts_df['CreatedDate'] = contacts_df['CreatedDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['CreatedDate'] = contacts_df['CreatedDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['LastModifiedDate'] = pd.to_datetime(contacts_df['LastModifiedDate'])
contacts_df['LastModifiedDate'] = contacts_df['LastModifiedDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['LastModifiedDate'] = contacts_df['LastModifiedDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['LastActivityDate'] = pd.to_datetime(contacts_df['LastActivityDate'])
contacts_df['LastActivityDate'] = contacts_df['LastActivityDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['LastActivityDate'] = contacts_df['LastActivityDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['LastCURequestDate'] = pd.to_datetime(contacts_df['LastCURequestDate'])
contacts_df['LastCURequestDate'] = contacts_df['LastCURequestDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['LastCURequestDate'] = contacts_df['LastCURequestDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['LastCUUpdateDate'] = pd.to_datetime(contacts_df['LastCUUpdateDate'])
contacts_df['LastCUUpdateDate'] = contacts_df['LastCUUpdateDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['LastCUUpdateDate'] = contacts_df['LastCUUpdateDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['Birthdate'] = pd.to_datetime(contacts_df['Birthdate'])
contacts_df['Birthdate'] = contacts_df['Birthdate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['Birthdate'] = contacts_df['Birthdate'].dt.strftime('%Y-%m-%d')
contacts_df['EmailBouncedDate'] = pd.to_datetime(contacts_df['EmailBouncedDate'])
contacts_df['EmailBouncedDate'] = contacts_df['EmailBouncedDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['EmailBouncedDate'] = contacts_df['EmailBouncedDate'].dt.strftime('%Y-%m-%d')
contacts_df['LastModifiedDate'] = pd.to_datetime(contacts_df['LastModifiedDate'])
contacts_df['LastModifiedDate'] = contacts_df['LastModifiedDate'].dt.tz_localize('UTC').dt.tz_convert('America/Los_Angeles')
contacts_df['LastModifiedDate'] = contacts_df['LastModifiedDate'].dt.strftime('%Y-%m-%d %I:%M:%S %p')
contacts_df['LastModifiedById'] = contacts_df['LastModifiedById'].str.lstrip('00530000000')
contacts_df['OwnerId'] = contacts_df['OwnerId'].str.lstrip('00530000000')
contacts_df['CreatedById'] = contacts_df['CreatedById'].str.lstrip('00530000000')
contacts_df['MasterRecordId'] = contacts_df['MasterRecordId'].str.lstrip('00Q')
contacts_df['ConvertedAccountId'] = contacts_df['ConvertedAccountId'].str.lstrip('00Q')
contacts_df['ConvertedContactId'] = contacts_df['ConvertedContactId'].str.lstrip('00Q')
contacts_df['ConvertedOpportunityId'] = contacts_df['ConvertedOpportunityId'].str.lstrip('00Q')
contacts_df['LastModifiedById'] = contacts_df['LastModifiedById'].str.lstrip('00Q')


import json
import os
import re
from enum import Enum, auto
from typing import List, Optional, Union
from dataclasses import dataclass

consistent_datatypes = {"numeric": ["bit", "smallmoney", "money", "int",
                                    "decimal", "decimal(4,2)", "decimal(3,2)", "float", "float64", "double",
                                    "bigdecimal", "int32", "int64", "bigint",
                                    "numeric", "bignumeric", "smallint", "integer",
                                    "tinyint", "byteint", "int8", "long",
                                    "float8", "hugeint", "int4", "signed",
                                    "real", "float4", "int2", "short",
                                    "int1", "ubigint", "uinteger", "usmallint",
                                    "utinyint", "bool", "boolean", "mediumint",
                                    "double precision", "dec", "smallserial", "serial",
                                    "bigserial", "number", "number(3,2)", "binary", "smalldecimal"],
                        "date":
                            ["date", "datetimeoffset", "datetime2", "smalldatetime",
                             "timestamp", "datetime", "time", "year",
                             "interval", "timestamp_ltz", "timestamp_ntz", "timestamp_tz"],
                        "string":
                            ["string", "varchar", "varchar2", "char",
                             "bpchar", "text", "nchar", "nvarchar",
                             "tinyblob", "blob", "shorttext", "clob",
                             "nclob", "ntext", "tinytext", "mediumtext",
                             "longtext", "mediumblob", "longblob", "image",
                             "character", "character varying", "binary", "varbinary"]}


def change_operator(operator):
    if operator == "GOE":
        operator = ">="
    elif operator == "LOE":
        operator = "<="
    elif operator == "EQ":
        operator = "=="
    elif operator == "GT":
        operator = ">"
    elif operator == "LT":
        operator = "<"
    return operator


class SQLFunction(Enum):
    MAX = auto()
    MIN = auto()
    AVG = auto()
    COUNT = auto()
    SUM = auto()


class TimeDimensionGranularity(Enum):
    year = auto()
    quarter = auto()
    month = auto()
    week = auto()
    day = auto()
    hour = auto()
    interval = auto()


class SignOperator(Enum):
    NEXT = auto()
    PREV = auto()


@dataclass
class Field:
    column: Optional[str] = None
    aggregation: Optional[Union[SQLFunction, TimeDimensionGranularity, str]] = None
    dataset: Optional[str] = None
    entityType: Optional[str] = None
    alias: Optional[str] = None
    complexExpression: Optional[str] = None
    internalDataType: Optional[str] = None
    sourceDataType: Optional[str] = None
    sqlExpression: Optional[str] = None
    sourceFormat: Optional[str] = None

    def __repr__(self):
        if self.column is not None:
            if self.aggregation is not None:
                if isinstance(self.aggregation, SQLFunction) or isinstance(self.aggregation,
                                                                            TimeDimensionGranularity):
                    field_with_agg = self.aggregation.name + " ( " + self.column + " )"
                else:
                    field_with_agg = self.aggregation + " ( " + self.column + " )"
            else:
                field_with_agg = self.column
        elif self.complexExpression is not None:
            field_with_agg = self.complexExpression
        if self.alias is not None:
            field_with_agg += " AS " + self.alias
        return field_with_agg


class SQLOperator(Enum):
    EQ = auto()
    NOT_EQ = auto()
    GOE = auto()
    GT = auto()
    LOE = auto()
    LT = auto()
    IN = auto()
    NOT_IN = auto()
    SEARCH = auto()
    TIMESPAN = auto()
    FOREACH = auto()


class PatternOperator(Enum):
    START = auto()
    END = auto()
    CONTAINS = auto()


class BooleanOperator(Enum):
    AND = auto()
    OR = auto()


class PeriodOperator(Enum):
    PREV = auto()
    CURR = auto()
    NEXT = auto()


class TimeOperator(Enum):
    RANGE = auto()
    FROM = auto()
    TO = auto()


@dataclass
class Condition:
    field: Field
    operator: Optional[Union[SQLOperator, TimeOperator, PatternOperator, str]] = None
    value: Optional[List[Union[float, str]]] = None
    type: Optional[str] = None
    steps: Optional[str] = None
    interval: Optional[str] = None
    direction: Optional[Union[PeriodOperator, str]] = None
    negate: Optional[bool] = None

    def __repr__(self):
        vars = []
        if self.value is not None:
            for var in self.value:
                vars.append(str(var))
            formatted_value = "( " + ", ".join(vars) + " )"
        else:
            formatted_value = None

        if isinstance(self.operator, str):
            operator = self.operator
        elif isinstance(self.operator, SQLOperator) or isinstance(self.operator, TimeOperator):
            operator = self.operator.name
            operator = change_operator(operator)
        else:
            operator = None
        if self.field.column is not None:
            field_with_agg = self.field.column
            if self.field.aggregation is not None:
                if isinstance(self.field.aggregation, SQLFunction) or isinstance(self.field.aggregation,
                                                                                 TimeDimensionGranularity):
                    field_with_agg = self.field.aggregation.name + " ( " + field_with_agg + " )"
                else:
                    field_with_agg = self.field.aggregation + " ( " + field_with_agg + " )"
        elif self.field.complexExpression is not None:
            field_with_agg = self.field.complexExpression

        if formatted_value is not None and self.direction is None:
            where_condition = (
                    field_with_agg + " " + operator + " " + str(formatted_value)
            )
        elif formatted_value is None and self.direction is not None:
            if isinstance(self.direction, PeriodOperator):
                direction = self.direction.name
            else:
                direction = self.direction
            if self.steps is not None and self.interval is not None:
                where_condition = (
                        field_with_agg + " " + operator + " " + self.interval + " " + direction + " " + self.steps
                )
            elif self.steps is not None and self.interval is None:
                where_condition = (
                        field_with_agg + " " + operator + " " + direction + " " + self.steps
                )
            elif self.steps is None and self.interval is not None:
                where_condition = (
                        field_with_agg + " " + operator + " " + self.interval + " " + direction
                )
            else:
                where_condition = (
                        field_with_agg + " " + operator + " " + direction
                )
        elif operator is not None:
            where_condition = (
                    field_with_agg + " " + operator
            )
        else:
            where_condition = (
                field_with_agg
            )

        return where_condition


@dataclass
class CompositeCondition:
    operator: Union[BooleanOperator, str]
    conditions: List[Union[Condition, 'CompositeCondition']]

    def __repr__(self):
        if isinstance(self.operator, BooleanOperator):
            where_condition = '( ' + (' ' + self.operator.name + ' ').join([str(c) for c in self.conditions]) + ' )'
        else:
            where_condition = '( ' + (' ' + self.operator + ' ').join([str(c) for c in self.conditions]) + ' )'

        return where_condition


class SQLSorting(Enum):
    DESC = auto()
    ASC = auto()


@dataclass
class Sorting:
    field: str
    order: Union[SQLSorting, str]

    def __repr__(self):
        if isinstance(self.order, SQLSorting):
            return self.field + " " + self.order.name
        return self.field + " " + self.order


@dataclass
class Component:
    type: str
    queryId: str


@dataclass
class QueryComponent(Component):
    category: str


@dataclass
class ChartComponent(Component):
    chartType: str


@dataclass
class MapComponent(Component):
    mapType: str


@dataclass
class From:
    dataset: str


@dataclass
class Join:
    query: str
    workspace: Optional[str] = None
    dataset: Optional[str] = None
    on: Optional[List[str]] = None
    left_on: Optional[List[str]] = None
    right_on: Optional[List[str]] = None


@dataclass
class GroupBy:
    field: Field


@dataclass
class Query:
    fields: List[Union[Field]]
    pivot: Optional[List[Field]] = None
    join: Optional[List[Join]] = None
    id: Optional[str] = None
    datasets: Optional[List[From]] = None
    where: Optional[List[Union[Condition, CompositeCondition]]] = None
    groupBy: Optional[List[GroupBy]] = None
    orderBy: Optional[List[Sorting]] = None
    limit: Optional[Union[int, str]] = None
    offset: Optional[Union[int, str]] = None

    def to_sql(self, dataset: str = None):
        sql = "SELECT {} FROM {}"

        fields_with_agg = [str(field) for field in self.fields]
        formatted_fields = ", ".join(fields_with_agg)

        if self.datasets is not None:
            froms_array = [d.dataset for d in self.datasets]
            table = ", ".join(froms_array)
        elif dataset is not None:
            table = dataset
        else:
            table = "{{dataset.A}}"
        sql = sql.format(formatted_fields, table)

        if self.where is not None:
            sql_where = " WHERE {}"

            where_conditions = [str(condition) for condition in self.where]
            if where_conditions:
                formatted_where_conditions = " AND ".join(where_conditions)
                sql_where = sql_where.format(formatted_where_conditions)
                sql += sql_where

        if self.orderBy is not None:
            sql_orderby = " ORDER BY {}"
            sorting_conditions = [str(sorting) for sorting in self.orderBy]
            formatted_sorting = ", ".join(sorting_conditions)
            sql_orderby = sql_orderby.format(formatted_sorting)
            sql += sql_orderby

        if self.limit is not None:
            sql += " LIMIT " + str(self.limit)

        return sql

    def _field_spell_out(self, field):
        # otherwise, check if internal and original data types are not consistent each other
        global consistent_datatypes
        consistent_vec = consistent_datatypes[field.internalDataType.lower()]
        if (
                field.internalDataType.lower() == "numeric" or field.internalDataType.lower() == "string") and field.sourceDataType.lower() not in consistent_vec:
            # if internal and original data types are not consistent each other use this pattern
            # aggregation column (original data type, to internal data type ) AS Alias
            if field.alias is None:
                if field.aggregation is not None and isinstance(field.aggregation, str):
                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
                else:
                    string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
            else:
                if field.aggregation is not None and isinstance(field.aggregation, str):
                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
                else:
                    string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
        elif field.internalDataType.lower() == "date" and field.sourceDataType.lower() not in consistent_vec:
            # if internal and original data types are not consistent each other use this pattern
            # aggregation column (original data type, to internal data type, date_format YYYY-MM-DD ) AS Alias
            if field.sourceFormat is not None:
                if field.alias is None:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
                    else:
                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
                else:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
                    else:
                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
            else:
                if field.alias is None:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
                    else:
                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
                else:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
                    else:
                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
        else:
            # else
            if field.alias is None:
                if field.aggregation is not None and isinstance(field.aggregation, str):
                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ")" + '"'
                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ")" + '"'
                else:
                    string = '"' + field.column + " (" + field.sourceDataType + ")" + '"'
            else:
                if field.aggregation is not None and isinstance(field.aggregation, str):
                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
                else:
                    string = '"' + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
        return string

    def _where_spell_out(self, where):
        global consistent_datatypes
        conditions = []

        for where_condition in where:
            if isinstance(where_condition, Condition):
                if where_condition.field.column is not None:
                    vars = []
                    if where_condition.value is not None:
                        for var in where_condition.value:
                            vars.append(str(var))
                        formatted_value = "( " + ", ".join(vars) + " )"
                    else:
                        formatted_value = None

                    if isinstance(where_condition.operator, str):
                        if where_condition.negate is None:
                            operator = where_condition.operator
                        else:
                            operator = "NOT " + where_condition.operator
                    elif isinstance(where_condition.operator, SQLOperator) or isinstance(where_condition.operator,
                                                                                         TimeOperator):
                        operator = where_condition.operator.name
                        if operator == "IN":
                            if where_condition.negate is None:
                                operator = where_condition.operator.name
                            else:
                                operator = "NOT " + where_condition.operator.name
                        elif operator == "GOE":
                            if where_condition.negate is None:
                                operator = ">="
                            else:
                                operator = "<"
                        elif operator == "LOE":
                            if where_condition.negate is None:
                                operator = "<="
                            else:
                                operator = ">"
                        elif operator == "EQ":
                            if where_condition.negate is None:
                                operator = "=="
                            else:
                                operator = "!="
                        elif operator == "GT":
                            if where_condition.negate is None:
                                operator = ">"
                            else:
                                operator = "<="
                        elif operator == "LT":
                            if where_condition.negate is None:
                                operator = "<"
                            else:
                                operator = ">="
                    else:
                        operator = None

                    consistent_vec = consistent_datatypes[where_condition.field.internalDataType.lower()]
                    if (
                            where_condition.field.internalDataType.lower() == "numeric" or where_condition.field.internalDataType.lower() == "string") \
                            and where_condition.field.sourceDataType.lower() not in consistent_vec:
                        # if internal and original data types are not consistent each other use this pattern
                        # aggregation column (original data type, to internal data type ) AS Alias
                        formatted_cond = where_condition.field.column + " (" + where_condition.field.sourceDataType + ", to " + where_condition.field.internalDataType + ") " + operator + " " + formatted_value
                    elif where_condition.field.internalDataType.lower() == "date" and where_condition.field.sourceDataType.lower() not in consistent_vec:
                        # if internal and original data types are not consistent each other use this pattern
                        # aggregation column (original data type, to internal data type, date_format YYYY-MM-DD ) AS Alias
                        if where_condition.field.sourceFormat is not None:
                            formatted_cond = where_condition.field.column + " (" + where_condition.field.sourceDataType + ", to " + where_condition.field.internalDataType + ", format " + where_condition.field.sourceFormat + ") " + operator + " " + formatted_value
                        else:
                            formatted_cond = where_condition.field.column + " (" + where_condition.field.sourceDataType + ", to " + where_condition.field.internalDataType + ", format YYYY-MM-DD" + ") " + operator + " " + formatted_value
                    else:
                        # else
                        formatted_cond = where_condition.field.column + " (" + where_condition.field.sourceDataType + ") " + operator + " " + formatted_value

                    conditions.append(formatted_cond)
                elif where_condition.field.complexExpression is not None:
                    if where_condition.field.alias is not None:
                        formatted_cond = where_condition.field.complexExpression + " AS " + where_condition.field.alias
                    else:
                        formatted_cond = where_condition.field.complexExpression
                    conditions.append(formatted_cond)
            elif isinstance(where_condition, CompositeCondition):
                other_conds = self._where_spell_out(where_condition.conditions)
                if where_condition.operator.name == "AND":
                    conditions = other_conds
                elif where_condition.operator.name == "OR":
                    conditions.append(" OR ".join(other_conds))
        return conditions

    def spell_out(self):
        spell_str = '# {"fields":['

        fields = []
        for field in self.fields:
            # if there is sql expression it always wins
            if field.sqlExpression is not None:
                if field.alias is None:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.sqlExpression + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.sqlExpression + '"'
                    else:
                        string = '"' + field.sqlExpression + '"'
                else:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.sqlExpression + " AS " + field.alias + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.sqlExpression + " AS " + field.alias + '"'
                    else:
                        string = '"' + field.sqlExpression + " AS " + field.alias + '"'
            elif field.complexExpression is not None:
                if field.alias is None:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.complexExpression + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.complexExpression + '"'
                    else:
                        string = '"' + field.complexExpression + '"'
                else:
                    if field.aggregation is not None and isinstance(field.aggregation, str):
                        string = '"' + field.aggregation + " " + field.complexExpression + " AS " + field.alias + '"'
                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
                        string = '"' + field.aggregation.name + " " + field.complexExpression + " AS " + field.alias + '"'
                    else:
                        string = '"' + field.complexExpression + " AS " + field.alias + '"'
            else:
                string = self._field_spell_out(field)
            fields.append(string)
        spell_str += ", ".join(fields) + "]"

        if self.datasets is not None and self.datasets:
            spell_str += ', "from":["{table}"]'

        if self.where is not None and self.where:
            spell_str += ', "where":['
            condition_str = self._where_spell_out(self.where)
            condition_str = ['"' + cond + '"' for cond in condition_str]
            spell_str += ", ".join(condition_str) + "]"

        if self.groupBy is not None and self.groupBy:
            spell_str += ', "group_by":['
            groupby_vec = []
            for clause in self.groupBy:
                groupby_vec.append(clause.field.column)
            spell_str += ", ".join(groupby_vec) + "]"

        if self.orderBy is not None and self.orderBy:
            spell_str += ', "order_by":['
            orderby_vec = []
            for clause in self.orderBy:
                if isinstance(clause.order, SQLSorting):
                    orderby_vec.append('"' + clause.field + " " + clause.order.name + '"')
                else:
                    orderby_vec.append('"' + clause.field + " " + clause.order + '"')
            spell_str += ", ".join(orderby_vec) + "]"

        if self.limit is not None:
            spell_str += ', "limit": ' + str(self.limit)

        if self.offset is not None:
            spell_str += ', "offset": ' + str(self.offset)
        else:
            spell_str += ', "offset": 0'

        spell_str += "}"
        return spell_str


@dataclass
class SmartQuery:
    queries: List[Query]
    components: Optional[List[Union[ChartComponent, MapComponent, QueryComponent, Component]]] = None
    javascript: Optional[List[str]] = None

    def spell_out(self):
        spell_array = []
        for query in self.queries:
            spell = query.spell_out()
            spell_array.append(spell)
        return spell_array

    @staticmethod
    def _get_tokens():
        f = open(os.path.join(os.path.dirname(__file__), "askdata_config", "compression_tokens.json"))
        comp_tokens = json.load(f)
        f.close()
        return comp_tokens

    @staticmethod
    def compress(smartquery_json: str):
        comp_tokens = SmartQuery._get_tokens()
        for token in comp_tokens:
            extended = token['decode']
            code = token['code']
            if extended in smartquery_json:
                smartquery_json = re.sub(r"\b" + extended + r"\b", code, smartquery_json)
        return smartquery_json

    @staticmethod
    def decompress(smartquery_json: str):
        comp_tokens = SmartQuery._get_tokens()
        for token in comp_tokens:
            code = token['code']
            decode = token['decode']
            if code in smartquery_json:
                smartquery_json = re.sub(r"\b" + code + r"\b", decode, smartquery_json)
        return smartquery_json



import json
from pandasql import sqldf
import pandas as pd

''' Integration abstract class '''
class Dataset():

	def discovery(self):
		pass

	def resultset_to_df(json_data):
		# Parse the JSON data
		json_data = json.loads(json_data)
		# Get the data from the JSON
		data = json_data['data']
		# Get the schema from the JSON
		schema = json_data['schema']
		# Create a list of column names
		column_names = [column['name'] for column in schema]
		# Create a list of data
		data_list = [row['cells'] for row in data]
		# Create a dataframe from the data
		df = pd.DataFrame(data_list, columns=column_names)
		# Return the dataframe
		return df

	def df_to_resultset(df):
		# Get the schema from the dataframe
		schema = df.columns
		# Create a list of dictionaries with the data
		data_list = df.to_dict('records')
		# Create a dictionary with the data
		data = {'data': data_list}
		# Create a dictionary with the schema
		schema_dict = {column: {'name': column} for column in schema}
		# Create a dictionary with the schema and data
		json_dict = {
			'schema': schema_dict,
			'data': data_list,
			'id': None,
			'dataset': {
				'id': '6512cf85-4f81-4fa0-a2b8-c51c6aad6cf9-BIGQUERY-a0fb8755-d50b-412b-9b86-0fd7d5529843',
				'name': 'Datacenters',
				'icon': 'https://storage.googleapis.com/askdata/datasets/icons/icoDataBigQuery.png'
			},
			'connection': '',
			'executedSQLQuery': 'SELECT AVG(`traffic`) AS `alias_0` FROM `askdata.Atlassian.atlassian_datacenter` ORDER BY `alias_0` DESC LIMIT 50',
			'schema': schema_dict,
			'data': data_list,
			'filters': None,
			'sortedBy': None,
			'page': None,
			'limit': 50,
			'totalRecords': 1,
			'queryId': ''
		}
		# Return the dictionary
		return json_dict

	def run_query(query, df):
		
		### Query is a SQLite3
		#query = """
		#SELECT *
		#FROM df
		#WHERE col_1 IS NOT NULL;
		#"""
		###

		# Run the query

		df_results = sqldf.run(query)
		return 	df_to_resultset(df_results)

	dataset_settings_example = """{
	    "payload": {
	        "page": 0,
	        "totalElements": 8,
	        "data": [
	            {
	                "datasetId": "f324ef9a-279b-4105-901b-c47c624edf7f-DATA_TABLE-0d129d26-58e5-441f-b9b3-8c80efdeb094",
	                "schemaMetaData": {
	                    "columnId": "area_geografica",
	                    "columnName": "area_geografica",
	                    "dataType": "VARCHAR",
	                    "dataExample": "Calabria",
	                    "internalDataType": "STRING",
	                    "indexedWith": null,
	                    "join": null,
	                    "details": {}
	                },
	                "parameterType": "ENTITY_TYPE",
	                "code": "AREA_GEOGRAFICA",
	                "name": "Area geografica",
	                "description": null,
	                "synonyms": [
	                    "area",
	                    "geografia",
	                    "area geografica",
	                    "area geograficas",
	                    "regione"
	                ],
	                "icon": "",
	                "sampleQueries": [],
	                "importValues": true,
	                "mandatory": false,
	                "enabled": true,
	                "advancedConfiguration": {},
	                "viewValues": "View",
	                "custom": false,
	                "dynamicParameterValues": [],
	                "searchable": false,
	                "nameTransformer": null,
	                "synonymTransformers": null
	            }
	        ]
	    }
	}"""

	### Convert string_json_object in a denormalized dataframe """
	def dataset_settings_to_dataframe(dataset_settings):
	    """
	    Convert string_json_object in a denormalized dataframe
	    """
	    # Convert string_json_object in a json object
	    json_object = json.loads(dataset_settings)
	    # Convert json_object in a dataframe
	    df = pd.DataFrame(json_object['payload']['data'])
	    # Return the dataframe
	    return df

### Convert string_json_object in a denormalized dataframe """
#df = dataset_settings_to_dataframe(dataset_settings_example)

### Show the dataframe
#df

	### Convert dataframe in a json similar to the string_json_object one """
	def dataset_settings_dataframe_to_json(df):
	    """
	    Convert dataframe in a json similar to the string_json_object one
	    """
	    # Convert dataframe in a json object
	    json_object = df.to_json(orient='records')
	    # Convert json_object in a string
	    string_json_object = json.dumps(json_object)
	    # Return the string
	    return string_json_object

### Convert dataframe in a json similar to the string_json_object one """
#dataset_settings = dataset_settings_dataframe_to_json(df)

### Show the string_json_object
#dataset_settings





from askdata import human2query
import pandas as pd

if __name__ == "__main__":
    df = pd.DataFrame()
    df = df.append({'player': 'Cristiano Ronaldo', 'goals': 10, 'team': 'Juventus'}, ignore_index=True)
    df = df.append({'player': 'Roberto Baggio', 'goals': 2, 'team': 'Milan'}, ignore_index=True)
    res = human2query.data2nl(df=df)
    print(res.head())



import requests
import pandas as pd

''' Create a function that get TIME_SERIES_DAILY_ADJUSTED for a particular symbol as input '''
def get_daily_adjusted(symbol):
    # create the endpoint
    url = f"https://www.alphavantage.co/query?function=TIME_SERIES_DAILY_ADJUSTED&symbol={symbol}&apikey={apikey}"
    # request the data
    response = requests.get(url)
    # parse the JSON data into a python dictionary
    data = response.json()
    # return the dictionary
    return data

''' Parse returned data in a denormalized pandas dataframe '''
def parse_daily_adjusted_in_a_dataframe(data):
    # extract the time series
    time_series = data["Time Series (Daily)"]
    # extract the columns to keep
    columns = ["5. adjusted close", "6. volume", "7. dividend amount", "8. split coefficient"]
    # create an empty list to store daily stock price
    rows = []
    # iterate over time_series
    for date, daily_prices in time_series.items():
        # create a dictionary for each date
        row = {key: value for key, value in daily_prices.items() if key in columns}
        # add a date column using the date string from the time_series
        row["date"] = date
        # add the symbol column
        row["symbol"] = data["Meta Data"]["2. Symbol"]
        # append the row to the rows list
        rows.append(row)
    # create a pandas dataframe from the rows list
    df = pd.DataFrame(rows)
    # reorder the columns
    df = df[["date", "symbol", "5. adjusted close", "6. volume", "7. dividend amount", "8. split coefficient"]]
    df.columns = ["Date", "Symbol", "Adjusted Close", "Volume", "Dividend amount", "Split coefficient"]
    # return the dataframe
    return df


''' 
Write a function that get as input a list of symbols and for earch symbol get get_daily_adjusted() and parse_daily_adjusted() returning a combination of the individual dataframes
'''

def get_daily_adjusted_df(symbols, apikey=None):

    if apikey=None:
        print("You need to request an Alpha Vantage API key from here: https://www.alphavantage.co/support/#api-key")

    # create an empty list to store dataframes
    dfs = []
    # iterate over symbols
    for symbol in symbols:
        # get the data for the symbol
        data = get_daily_adjusted(symbol, apikey)
        # parse the data into a dataframe
        df = parse_daily_adjusted_in_a_dataframe(data)
        # add the dataframe to the list
        dfs.append(df)
    # concatenate the list of dataframes
    df = pd.concat(dfs)

    # Casting the Volume type as float
    df["Volume"] = df["Volume"].astype(float)

    # return the concatenated dataframe
    return df


import requests
import yaml
import os
import pandas as pd
import logging
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)


class Channel:

    def __init__(self, env, token, _agentId, _domain):

        data = {
            "agent_id": _agentId
        }

        self._headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + token
        }

        if env == 'dev':
            self._base_url_ch = url_list['BASE_URL_FEED_DEV']
        if env == 'qa':
            self._base_url_ch = url_list['BASE_URL_FEED_QA']
        if env == 'prod':
            self._base_url_ch = url_list['BASE_URL_FEED_PROD']

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = self._base_url_ch + '/' + self._domain + '/agent/switch'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()
        self.r = r

    def load_channels(self):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels/'+'?agentId='+self._agentId+'&page=0&limit=100000'
        r = s.get(url=authentication_url, headers=self._headers)
        r.raise_for_status()
        df_channels = pd.DataFrame(r.json())

        return df_channels

    def create_channel(self, name, icon='https://s3.eu-central-1.amazonaws.com/innaas.smartfeed/icons/groupama/icone/channel/icon_channel_dm.png',
                       visibility='PRIVATE'):

        data = {
            "name": name,
            "icon": icon,
            "agentId": self._agentId,
            "visibility": visibility
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels/'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()
        return r.json()['id']

    def update_channel(self, channel_id, visibility,
                       icon='https://s3.eu-central-1.amazonaws.com/innaas.smartfeed/icons/groupama/icone/channel/icon_channel_dm.png',
                       iconFlag = False):
        #visibility is PUBLIC or PRIVATE
        if iconFlag:
            data = {
                "icon": icon,
                "visibility": visibility
            }
        else:
            data = {
                "visibility": visibility
            }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels/' + channel_id
        r = s.put(url=authentication_url, headers=self._headers,  json=data)
        r.raise_for_status()
        return r

    def delete_channel(self, channel_id):
        authentication_url = self._base_url_ch + '/channels/' + channel_id
        r = requests.delete(url=authentication_url, headers=self._headers)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            if r.status_code != 500:
                raise ex
            logging.info(ex)
            logging.info('Channel already deleted or not exist')

        return r

    def load_users_fromch(self, channel_id):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels/' + channel_id + '/users'
        r = s.get(url=authentication_url, headers=self._headers)

        r.raise_for_status()
        df_users = pd.DataFrame(r.json())
        return df_users

    def add_user_toch(self, channel_id, user_id, role="follower", mute="none"):

        data = {
            "userId": user_id,
            "role": role,
            "mute": mute
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_ch + '/channels/' + channel_id + '/users'
        r = s.post(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()
        return r

    def delete_user_fromch(self, channel_id, user_id):
        authentication_url = self._base_url_ch + '/channels/' + channel_id + '/users/' + user_id
        r = requests.delete(url=authentication_url, headers=self._headers)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            # usare libreria di logging
            if r.status_code != 409:
                raise ex
            logging.info(ex)
            logging.info('User already deleted or not exist')
        return r

    def un_mute_channel(self, channel_id):

        authentication_url = self._base_url_ch + '/channels/' + channel_id + '/unmute'
        r = requests.put(url=authentication_url, headers=self._headers)
        r.raise_for_status()
        return r

    def mute_channel(self, channel_id):

        data = {

            "period": "DAYS_7"
        }
        authentication_url = self._base_url_ch + '/channels/' + channel_id + '/mute'
        r = requests.put(url=authentication_url, headers=self._headers, json=data)
        r.raise_for_status()
        return r



import requests
from bs4 import BeautifulSoup
from itertools import product
from integrations.integration import Integration
import pandas as pd

''' get wikipedia page using beaufulsap that will be used to parse a wikipedia table '''
def get_wikipedia_page(url):
    response = requests.get(url)
    soup = BeautifulSoup(response.text, 'html.parser')
    return soup

''' get wikipedia table using beaufulsap that will be used to parse a wikipedia table '''
def get_wikipedia_table(soup):
    table = soup.find('table', {'class': 'wikitable'})
    return table

''' get wikipedia table using beaufulsap that will be used to parse a wikipedia table '''
def get_wikipedia_tables(soup):
    tables = soup.find_all('table', {'class': 'wikitable'})
    return tables

''' get wikipedia table headers using beaufulsap that will be used to parse a wikipedia table '''
def get_wikipedia_table_headers(table_rows):
    headers = []
    for th in table_rows[0].find_all('th'):
        headers.append(th.text.strip())
    return headers

def wikipedia_table_to_dataframe(table_tag):
    rowspans = []  # track pending rowspans
    rows = table_tag.find_all('tr')

    # first scan, see how many columns we need
    colcount = 0
    for r, row in enumerate(rows):
        cells = row.find_all(['td', 'th'], recursive=False)
        # count columns (including spanned).
        # add active rowspans from preceding rows
        # we *ignore* the colspan value on the last cell, to prevent
        # creating 'phantom' columns with no actual cells, only extended
        # colspans. This is achieved by hardcoding the last cell width as 1. 
        # a colspan of 0 means “fill until the end” but can really only apply
        # to the last cell; ignore it elsewhere. 
        colcount = max(
            colcount,
            sum(int(c.get('colspan', 1)) or 1 for c in cells[:-1]) + len(cells[-1:]) + len(rowspans))
        # update rowspan bookkeeping; 0 is a span to the bottom. 
        rowspans += [int(c.get('rowspan', 1)) or len(rows) - r for c in cells]
        rowspans = [s - 1 for s in rowspans if s > 1]

    # it doesn't matter if there are still rowspan numbers 'active'; no extra
    # rows to show in the table means the larger than 1 rowspan numbers in the
    # last table row are ignored.

    # build an empty matrix for all possible cells
    table = [[None] * colcount for row in rows]

    # fill matrix from row data
    rowspans = {}  # track pending rowspans, column number mapping to count
    for row, row_elem in enumerate(rows):
        span_offset = 0  # how many columns are skipped due to row and colspans 
        for col, cell in enumerate(row_elem.find_all(['td', 'th'], recursive=False)):
            # adjust for preceding row and colspans
            col += span_offset
            while rowspans.get(col, 0):
                span_offset += 1
                col += 1

            # fill table data
            rowspan = rowspans[col] = int(cell.get('rowspan', 1)) or len(rows) - row
            colspan = int(cell.get('colspan', 1)) or colcount - col
            # next column is offset by the colspan
            span_offset += colspan - 1
            value = cell.get_text().strip()
            for drow, dcol in product(range(rowspan), range(colspan)):
                try:
                    table[row + drow][col + dcol] = value
                    rowspans[col + dcol] = rowspan
                except IndexError:
                    # rowspan or colspan outside the confines of the table
                    pass

        # update rowspan bookkeeping
        rowspans = {c: s - 1 for c, s in rowspans.items() if s > 1}
    df = pd.DataFrame(table)

    header = df.iloc[0] #grab the first row for the header
    df = df[1:] #take the data less the header row
    df.columns = header 
    return df

class wikipedia_table(Integration):
    def load(self):
    	print('Mamma says hi')

#Test
url = 'https://en.wikipedia.org/wiki/2021%E2%80%9322_UEFA_Champions_League'
soup = get_wikipedia_page(url)
table = get_wikipedia_table(soup)
wikipedia_table_to_dataframe(table)

#Concatenated all the df from the same table
wikipedia_table_to_dataframe(get_wikipedia_tables(soup)[0]).append(wikipedia_table_to_dataframe(get_wikipedia_tables(soup)[1])).append(wikipedia_table_to_dataframe(get_wikipedia_tables(soup)[2]))


import pandas as pd

import requests

from oauth2client.service_account import ServiceAccountCredentials

from googleapiclient.discovery import build
from requests.adapters import HTTPAdapter
from urllib3 import Retry

import json


""" 
Read google sheet as dataframe


If you read with API, you need to use spreadsheetId and credentials (it is the json file in same notebook folder).
you also need to add the account email associated to the credentials as user shared in the google sheet.

"""
def read_gsheet(spreadsheetId, credentials = None, sheet = None, 

                valueRenderOption = 'FORMATTED_VALUE', dateTimeRenderOption = 'SERIAL_NUMBER'):
    

    if credentials is None:

        post_url = "https://api.askdata.com/smartnotebook/readGsheet"
        
        body = {
            "spreadsheetId": spreadsheetId,
            "sheet": sheet,
            "valueRenderOption": valueRenderOption,
            "dateTimeRenderOption": dateTimeRenderOption
        }
        
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        headers = {
            "Content-Type": "application/json"
        }
        r = s.post(url=post_url, json=body, headers=headers)
        r.raise_for_status()
        
        df = pd.DataFrame(json.loads(r.json()))

    else: 

        scope = ['https://www.googleapis.com/auth/spreadsheets.readonly']

        cred = ServiceAccountCredentials.from_json_keyfile_dict(credentials, scope)
        
        service = build('sheets', 'v4', credentials=cred)
    
        sheet_service = service.spreadsheets()
        
        range = 'A1:WWW500000'
        
        if sheet is not None:
            
            range = '%s!%s'%(sheet, range)
        
        result_input = sheet_service.values().get(spreadsheetId = spreadsheetId,
                                          range = range,
                                          valueRenderOption = valueRenderOption,
                                          dateTimeRenderOption = dateTimeRenderOption).execute()
        
        values_input = result_input.get('values', [])
        
        df = pd.DataFrame(values_input[1:], columns=values_input[0])

    return df


""" 
Read hubspot contacts as dataframe
Usage example: read_hubspot_contacts(api_key)
"""


def read_hubspot_contacts(api_key, offset=100):
    from askdata.integrations import hubspot
    return hubspot.get_contacts_df(api_key, offset)


""" 
Read alpha vantage api
Usage example: read_alphavantage_stock(api_key, symbols)
"""


def read_alphavantage_stock(symbols, api_key):
    from askdata.integrations import alphavantage
    return alphavantage.get_daily_adjusted_df(symbols, api_key)


def normalize_columns(df: pd.DataFrame):
    problematicChars = [",", ";", ":", "{", "}", "(", ")", "=", ">", "<", ".", "!", "?"]
    new_cols = {}
    for column in df.columns:

        columnName = column.lower()

        for p_char in problematicChars:
            columnName = columnName.replace(p_char, "")

        columnName = columnName.replace(" ", "_")
        columnName = columnName.replace("-", "_")
        columnName = columnName.replace("%", "perc")
        columnName = columnName.replace("+", "plus")
        columnName = columnName.replace("&", "and")
        columnName = columnName.replace("cross", "cross_pass")
        columnName = columnName.replace("authorization", "authoriz")
        columnName = columnName.strip()

        for p_char in problematicChars:
            columnName = columnName.replace(p_char, "")

        new_cols[column] = columnName

    return df.rename(columns=new_cols)


def read(type, settings):
    if type == "CSV":
        return __read_csv(settings)
    if type == "EXCEL":
        return __read_excel(settings)
    if type == "PARQUET":
        return __read_parquet(settings)
    if type == "GOOGLE_SHEETS":
        return __read_gsheet(settings)
    if type == "Hubspot":
        return __read_hubspot(settings)
    else:
        raise TypeError("Dataset type not supported yet")


def __read_csv(settings: dict):
    
    # Handle thousands
    if settings["thousands"] == "None":
       settings["thousands"] = None

    # Read source file
    df = pd.read_csv(filepath_or_buffer=settings["path"], sep=settings["separator"], encoding=settings["encoding"], thousands=settings["thousands"])

    # Detect if any column is a date-time
    for col in df.columns:
        if df[col].dtype == 'object':
            try:
                df[col] = pd.to_datetime(df[col])
            except ValueError:
                pass

    # Exec custom post-processing
    if "processing" in settings and settings["processing"] != "" and settings["processing"] != None:
        exec(settings["processing"])

    return df

def __read_parquet(settings: dict):
    
    df = pd.read_parquet(path=settings["path"])

    return df

def __read_excel(settings: dict):
    df = pd.read_excel(settings["path"])

    # Detect if any column is a date-time
    for col in df.columns:
        if df[col].dtype == 'object':
            try:
                df[col] = pd.to_datetime(df[col])
            except ValueError:
                pass

    return df

def __read_gsheet(settings: dict):
    df = read_gsheet(spreadsheetId=settings["spreadsheetId"], credentials=settings["credentials"], sheet=settings["sheet"],
    
                    valueRenderOption = settings['valueRenderOption'],  dateTimeRenderOption = settings['dateTimeRenderOption'])
                    
    return df

def __read_hubspot(settings: dict):
    df = read_hubspot_contacts(settings["fields"]["token"])

    return df




##################################
########## FACEBOOK ADS ##########
##################################


def read_fb_ads(my_app_id, my_app_secret, my_access_token, start_date, end_date, account_name):
    
    from askdata.integrations import facebook_api
    
    fb_ads = facebook_api.get_fb_ads(my_app_id, my_app_secret, my_access_token, start_date, end_date, account_name)

    return fb_ads



#################################
########## GOOGLE ADS ##########
################################


def read_google_ads(start_date, end_date, admin_account,
                   
                   yaml_file_path = 'googleads.yaml'):
    
    
    from askdata.integrations import google_api
    
    google_ads = google_api.get_google_ads(start_date, end_date, admin_account, yaml_file_path)
    
    return google_ads
    
    
    
    
###################################
########## AIRCALL CALLS ##########
###################################



def read_aircall_calls(token, api_key, #mandatory
                      
                      start = 7, end = 0 #optional
                     ):
    
    from askdata.integrations import aircall_api
    
    aircall_calls = aircall_api.get_aircall_calls(token, api_key, start, end)
    
    return aircall_calls
    
    
    
#############################
########## HUBSPOT ##########
#############################
    

## hubspot deals

def read_hubspot_deals(API_KEY, start_date, end_date):
    
    
    from askdata.integrations import hubspot_api
    
    deals_all_fiels = hubspot_api.get_hubspot_deals_all_fields(API_KEY, start_date, end_date)
    
    return deals_all_fiels
 

    
## hubspot contact

def read_hubspot_contact(API_KEY, start_date, end_date):
    
    from askdata.integrations import hubspot_api
    
    hubspot_contact = hubspot_api.get_hubspot_contact(API_KEY, start_date, end_date)
    
    return hubspot_contact
                         
           
        
## hubspot pipelines

def read_hubspot_pipelines(API_KEY):
    
    
    from askdata.integrations import hubspot_api
    
    hubspot_pipelines = hubspot_api.get_hubspot_pipelines(API_KEY)
    
    return hubspot_pipelines
                         
                         
    
## hubspot users

def read_hubspot_users(API_KEY):
    
    from askdata.integrations import hubspot_api
    
    hubspot_users = hubspot_api.get_hubspot_users(API_KEY)
    
    return hubspot_users
    


import requests
import yaml
import os
import logging
import pandas as pd
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from IPython.display import Image, display, HTML

_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

root_dir = os.path.abspath(os.path.dirname(__file__))
# retrieving base url
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to Python the dictionary format
    url_list = yaml.load(file, Loader=yaml.FullLoader)

class Insight_Definition:


    def __init__(self, env:str, token, agent_slug, definition):

        self.definition_id = definition["id"]
        self.agent_id = definition["agentId"]
        self.agent_slug = agent_slug
        self.collection_id = definition["collectionId"]
        self.name = definition["name"]
        self.slug = definition["slug"]
        self.icon = definition["icon"]
        self.components: list = definition["components"]
        self._env = env
        self._token = token

        if env.lower() == 'dev':
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_DEV']
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_DEV']
            self.app_url = "https://app-dev.askdata.com"
        if env.lower() == 'qa':
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_QA']
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_QA']
            self.app_url = "https://app-qa.askdata.com"
        if env.lower() == 'prod':
            self._base_url_askdata = url_list['BASE_URL_ASKDATA_PROD']
            self.smart_insight_url = url_list['BASE_URL_INSIGHT_PROD']
            self.app_url = "https://app.askdata.com"

    def __make_get_request(self, url):
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        r = s.get(url=url, headers=headers)
        r.raise_for_status()
        return r

    def __make_put_request(self, url, body):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, json=body, headers=headers)
        r.raise_for_status()
        return r

    def __make_post_request(self, url, body):
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        r = s.post(url=url, json=body, headers=headers)
        r.raise_for_status()
        return r

    def add_skin(self, skin_code:str):
        body = {

            "name": self.name,
            "slug": self.slug,
            "ui": {
                "skin": skin_code
            }
        }

        url = self.smart_insight_url+"/definitions/"+self.definition_id
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_widget(self, widget_type: str, query_ids:list=[], settings:dict={}, values:dict={}, texts:dict={}, charts:list=[]):
        position = (len(self.components))
        widget_id = self.__create_widget(position, widget_type)
        if(query_ids!=[]):
            self.edit_widget(widget_id, widget_type,query_ids, settings, values, texts, charts)
        return widget_id

    def edit_widget(self, widget_id, widget_type: str, query_ids:list, field_codes:dict, values:dict={}, texts:dict={}, charts:list=[]):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/widgets/"+widget_id
        columns = []
        for query in query_ids:
            if query.startswith("qcomp"):

                qc = self.get_component(query)
                url_preview = self.smart_insight_url + "/composed_queries/" + qc["settings"]["qc"] + "/preview?limit=100&agentId=" + self.agent_id
                r = self.__make_post_request(url_preview, {})
                cols = r.json()["schema"]
                for col in cols:
                    if(query not in ((str)(col["id"]))):
                        col["id"] = query + "-" +col["id"]
                    columns.append(col)
                #cols = self.get_columns(query_ids)
            else:
                cols = self.get_columns(query_ids)
                for col in cols:
                    columns.append(col)

        settings = {}
        for field in field_codes.keys():
            for column in columns:
                if(column["code"] in field_codes[field] or column["name"] in field_codes[field]):
                    for query in query_ids:
                        if(str(field_codes[field]).startswith(query)):
                           if(str(column["id"]).startswith(query)):
                               settings[field + "_id"] = column["id"]
                               break

                           else:
                               continue
                        else:
                            settings[field + "_id"] = column["id"]
                            break

        for key in values.keys():
            settings[key + "_id"] = values[key]

        body = {
            "id" : widget_id,
            "type": "widget",
            "name": "Widget",
            "customName": False,
            "dependsOn": [],
            "widgetType": widget_type,
            "queryIds": query_ids,
            "settings": settings,
            "texts": texts,
            "charts": charts,
            "valid": True,
            "queryComponent": False
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def get_columns(self, query_ids) -> list:

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/queries/columns"
        body = {"queryIds": query_ids}
        r = self.__make_post_request(url, body)
        return r.json()

    def display_card_img(self):
        display(Image(data=self.__get_card_img()))


    def download_card_img(self, image_path:str):
        with open(image_path, 'wb') as f:
            f.write(self.__get_card_img())


    def get_card_url(self):
        return self.__get_card__url()


    def __get_card__url(self):

        url = self.app_url + "/"+self.agent_slug+"/card/" + self.slug
        return url

    def preview(self):

        url = self.app_url+"/card/"+self.__get_card_id()
        display(HTML('<iframe src="'+url+'" height="300" width="300" title="'+self.slug+'"></iframe>'))

    def __get_card_id(self):

        url = self._base_url_askdata+"/smartfeed/agents/"+self.agent_id+"/insight-slugs/"+self.slug
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.get(url=url, headers=headers)
        r.raise_for_status()
        cards = r.json()
        for card in cards:
            try:
                if(card["insightSlug"]==self.slug):
                    return card["id"]
            except KeyError:
                pass

        logging.error("Card not found")
        return None

    def __get_card_img(self):

        card_id = self.__get_card_id()

        if(card_id != None):

            url = self._base_url_askdata+"/smartinsight/insights/"+card_id+"/data?type=image"
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            r = s.get(url=url, headers=headers)
            r.raise_for_status()
            return r.content
        else:
            return None


    def __create_widget(self, position, widget_type):

        body = {"type": "widget", "position": position, "widgetType": widget_type}
        url = self.smart_insight_url + '/definitions/' + self.definition_id + '/components/'

        r = self.__make_post_request(url, body)
        self.components = r.json()["components"]
        return self.components[position]["id"]

    def add_table(self, query="", columns=[], pivot_col=[], pivot_type="2", max_results=50):

        position = (len(self.components))
        body = {"type": "table", "position": position}

        url = self.smart_insight_url + '/definitions/' + self.definition_id + '/components/'
        r = self.__make_post_request(url, body)
        self.components = r.json()["components"]

        if(query != "" and columns!=[]):
            self.edit_table(self.components[position]["id"], query, columns, pivot_col, pivot_type, max_results)

        return self.components[position]["id"]

    def edit_table(self, table_id, query="", columns=[], pivot_col=[], pivot_type="2", max_results=50):

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/tables/" + table_id

        body = {
            "id": table_id,
            "type": "table",
            "name": "Table",
            "customName":False,
            "dependsOn": query,
            "queryId": query,
            "columns": columns,
            "maxResults": max_results,
            "valid": True,
            "queryComponent":False,
            "pivot": True,
            "pivotColumns": pivot_col,
            "pivotType": pivot_type
        }

        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def pivot_table(self, table_id, query="", columns=[], pivot_col=[], max_results=50):

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/tables/" + table_id

        body = {
            "id": table_id,
            "type": "table",
            "name": "Table",
            "customName": False,
            "dependsOn": [query],
            "queryId": query,
            "columns": columns,
            "maxResults": max_results,
            "valid": True,
            "queryComponent": False,
            "pivot": True,
            "pivotColumns": pivot_col,
            "pivotType": "2"
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, json=body, headers=headers)
        r.raise_for_status()
        self.components = r.json()["components"]

    def get_result_set(self, query_id):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/queries/resultset/"+query_id

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.get(url=url, headers=headers)
        r.raise_for_status()
        response = r.json()
        columns = []
        for column in response["schema"]:
            columns.append(column["name"])

        result_set = {}

        for i in range(len(columns)):
            row = []
            for temp_row in response["data"]:
                row.append(temp_row["cells"][i]["rawValue"])
            result_set[columns[i]] = row

        return pd.DataFrame(data=result_set)


    def add_chart(self, type:str, query:str, params:list, title = "Chart"):
        position = (len(self.components))

        chart_id = self.add_component("chart", position)

        self.edit_chart(chart_id, type, query, params, title)


        return self.components[position]["id"]


    def edit_chart(self, chart_id, chart_type:str, query:str, params, title = "Chart"):

        if(not query.startswith("qcomp")):

            url = self._base_url_askdata+"/smartinsight/definitions/"+self.definition_id+"/queries/"+query.lower()+"/columns"
            r = self.__make_get_request(url)
            columns = r.json()

            while(columns == []):

                r = self.__make_get_request(url)
                columns = r.json()

        else:
            qc = self.get_component(query)

            url_preview = self.smart_insight_url+"/composed_queries/"+qc["settings"]["qc"]+"/preview?limit=100&agentId="+self.agent_id

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {"Authorization": "Bearer" + " " + self._token}
            response = s.post(url=url_preview, json={}, headers=headers)
            response.raise_for_status()
            r = response.json()
            columns = r["schema"]

        upper_params = []
        for param in params:
            for column in columns:
                if((param in column["code"]) or (param == column["code"])):
                    upper_params.append(column["code"])

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/charts/" + chart_id


        body = {
            "chartType": chart_type.upper(),
            "customName": False,
            "dependsOn": [],
            "id": chart_id,
            "name": title,
            "params": upper_params,
            "queryComponent": False,
            "queryId": query,
            "type": "chart",
            "valid": True
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]


    def add_list(self, query_id, title_parameter, description_parameter, action_enabled, action_label,
                  action_url, max_results,list_name="List"):
        position = (len(self.components))
        list_id = self.add_component(type="list", position=position)
        self.edit_list(list_id, query_id, title_parameter,
                       description_parameter, action_enabled, action_label, action_url, max_results, list_name)
        return list_id


    def edit_list(self, list_id, query_id, title_parameter, description_parameter, action_enabled, action_label,
                  action_url, max_results,list_name="List"):

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/lists/" + list_id
        body = {
            "id":list_id,
            "type":"list",
            "name":list_name,
            "customName":True,
            "dependsOn":[],
            "queryId":query_id,
            "title":title_parameter,
            "description":description_parameter,
            "actionEnabled": action_enabled,
            "actionLabel": action_label,
            "actionUrl":action_url,
            "maxResults":max_results,
            "valid":True,
            "queryComponent":False
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_data(self, df:pd.DataFrame, parameters_settings:list=[]):
        position = (len(self.components))
        data_id = self.add_component("data", position)
        self.edit_data(data_id, df)
        if(parameters_settings!=[]):
            for i in range(len(parameters_settings)):
               self.edit_data_entity(data_id, i, parameters_settings[i])
        return data_id


    def add_join_to_qc(self, qc_id, new_dataset_slug, new_fields, left_on:list, right_on:list, type:str ,filters:list=[], sorted_by= None, limit=50, offset=0, pivot=[]):

        new_qc = self.create_qc(new_dataset_slug, new_fields, filters, sorted_by, limit= limit, pivot=pivot, offset=offset)

        url = self.smart_insight_url+"/composed_queries/{}".format(qc_id)

        qc = self.get_qc(qc_id)
        qc = qc["settings"]["qcSettings"]
        qc["joinedSmartQuery"]= {
            "smartQueryId": new_qc["id"]
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, json=qc, headers=headers)
        r.raise_for_status()
        response = r.json()

        right_fields = []
        left_fields = []
        for field in new_qc["fields"]:
            if field["column"] in right_on:
                right_fields.append(field)
        for field in qc["fields"]:
            if field["column"] in left_on:
                left_fields.append(field)

        joinable_fields = response["joinedSmartQuery"]["joinableFields"]

        qc["joinedSmartQuery"]["smartQueryId"] = new_qc["id"]
        qc["joinedSmartQuery"]["joinType"] = type.upper()
        qc["joinedSmartQuery"]["leftFields"] = left_fields
        qc["joinedSmartQuery"]["rightFields"] = right_fields
        qc["joinedSmartQuery"]["joinableFields"] = joinable_fields
        qc["joinedSmartQuery"]["dataset"] = {
            "id": new_qc["datasets"][0]["dataset"],
            "slug": new_dataset_slug
        }


        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, json=qc, headers=headers)
        r.raise_for_status()
        response = r.json()



    def get_qc(self, qc_id):
        for component in self.components:
            if component["id"].startswith("qcomp"):
                if component["settings"]["qc"] == qc_id:
                    return component

    def add_query_composer(self, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50, offset=0, pivot=[]):

        qc_id = self.create_qc(dataset_slug, fields, filters, sorted_by, dataset_id, limit, offset, pivot)["id"]

        position = (len(self.components))
        qc_post_url = self.smart_insight_url + "/definitions/" + self.definition_id + "/query_composers"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        body = {"type": "query_composer", "position": position, "qc": qc_id}

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=qc_post_url, json=body, headers=headers)
        response.raise_for_status()
        query = response.json()
        self.components = query["components"]

        return {
            "qc_id": qc_id,
            "component_id": query["components"][position]["id"]
        }


    def create_qc(self, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50, offset=0, pivot=[]):
        """
        Pivot columns MUST be contained in field list
        """
        if(fields==[]):
            print("SELECT FIELDS")
            return

        if sorted_by is None:
            sorted_by = []

        if filters is None:
            filters = []

        if(dataset_id=="" and dataset_slug!=""):
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            url = self._base_url_askdata + '/smartbot/agents/agentslug/' + self.agent_slug + '/datasetslug/' + dataset_slug

            headers = {"Authorization": "Bearer" + " " + self._token}
            response = s.get(url=url, headers=headers)
            response.raise_for_status()
            r = response.json()

            datasetId = r["datasetId"]

        elif(dataset_id!=""):
            datasetId=dataset_id
        else:
            print("SPECIFY A DATASET ID OR SLUG")
            return

        body = {
            "agentId": self.agent_id,
            "datasetId": datasetId,
            "fields": fields,
            "sortedBy": sorted_by,
            "filters": filters,
            "limit": limit,
            "offset": offset,
            "pivot": pivot
        }

        url = self.smart_insight_url + "/composed_queries/sdk"

        response = self.__make_post_request(url, body)

        return response.json()


    def get_component(self, component_id):
        for component in self.components:
            if (component["id"] == component_id):
                return component


    def add_query_composer_to_data_component(self, data_id, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50, offset=0, pivot=[]):

        qc_id = self.create_qc(dataset_slug, fields, filters, sorted_by, dataset_id, limit, offset, pivot)["id"]
        name = self.get_component(data_id)["settings"]["name"]
        variable_name = self.get_component(data_id)["settings"]["variableName"]
        body = {
            "name": name,
            "variableName": variable_name,
            "qcId": qc_id
        }
        url = self._base_url_askdata+"/smartinsight/definitions/"+self.definition_id+"/data_queries/"+data_id
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def edit_data(self, data_id: str, df: pd.DataFrame, index:bool=False, qc_id=None):

        body = df.to_json(orient="table", index=index)
        if(qc_id!=None):
            body["qcId"] = qc_id

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/data_queries/"+data_id+"/data"
        #logging.info("URL: {}".format(url))
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, data=body, headers=headers)
        r.raise_for_status()
        self.components = r.json()["components"]


    def load_data_entities(self, data_id):
        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/queries"
        #logging.info("URL {}".format(url))
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url, headers=headers)
        response.raise_for_status()
        r = response.json()
        for query in r:
            if(query["id"]==data_id):
                query_data = query
                return query_data["entities"]
        logging.info("No query with id {} found".format(data_id))
        return []

    #TODO Metodo per ottenere una specifica entity e relativa posizione

    def edit_data_entity(self, data_id, entity_position:int, entity_settings:dict):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/queries/"+data_id+"/columns/"+str(entity_position)

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        r = s.put(url=url, json=entity_settings, headers=headers)
        r.raise_for_status()

    def add_to_channel(self, new_channel_id):
        url = self._base_url_askdata + "/smartfeed/pin/pinned/"+new_channel_id+"?addMode=true"
        body = [
            {
                "targetId": self.definition_id,
                "type": "CARD",
                "pinnedValues": {}
            }
        ]
        self.__make_put_request(url, body)

    def edit_data_name(self, data_id, component_name):
        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/data_queries/"+data_id
        body = {
            "name": component_name
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_cta(self, cta_label, cta_url):
        position = (len(self.components))
        cta_id = self.add_component("cta", position)
        self.edit_cta(cta_id, cta_label, cta_url)
        return cta_id

    def edit_cta(self, id, label, cta_url):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/ctas/"+id
        body = {
            "id":id,
            "type":"cta",
            "name":"Cta 1",
            "customName":False,
            "dependsOn":[],
            "valid":True,
            "queryComponent":False,
            "label":label,
            "link":cta_url
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_compare(self, name, second_item_enabled, query_id_1, measure_1, label_1, description_1,
                     query_id_2, measure_2, label_2, description_2, formatter_type="HIGH"):
        position = (len(self.components))
        compare_id = self.add_component("comparison", position)
        self.edit_compare(compare_id, name, second_item_enabled, query_id_1, measure_1, label_1, description_1,
                     query_id_2, measure_2, label_2, description_2, formatter_type)
        return compare_id

    def edit_compare(self, compare_id, name, second_item_enabled, query_id_1, measure_1, label_1, description_1,
                     query_id_2, measure_2, label_2, description_2, formatter_type="HIGH"):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/comparisons/"+compare_id
        body = {
            "name": name,
            "itemA": {
                "queryId": query_id_1,
                "measure":measure_1,
                "label":label_1,
                "description":description_1,
                "measureComparator":None
            },
            "itemB": {
                "queryId":query_id_2,
                "measure":measure_2,
                "label":label_2,
                "description": description_2,
                "measureComparator":None
            },
            "secondItemEnabled": second_item_enabled,
            "formatterType":formatter_type
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]


    def add_text(self, text):
        position = (len(self.components))
        text_id = self.add_component("text", position)
        self.edit_text(text_id, text)
        return text_id


    def edit_text(self, text_id , text, name="Text"):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/texts/"+text_id

        body = {
            "customName": False,
            "dependsOn": [],
            "id": text_id,
            "name": name,
            "queryComponent": False,
            "text": text,
            "type": "text",
            "valid": True
        }

        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]


    def add_html(self, code):
        position = (len(self.components))
        html_id = self.add_component("html", position)
        self.edit_html(html_id, code)
        return html_id


    def edit_html(self, html_id, code):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/htmls/"+html_id+"/content"

        body = {
            "html": code
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]


    def add_script(self, script):
        position = (len(self.components))
        script_id = self.add_component("script", position)
        self.edit_script(script_id, script)
        return script_id


    def edit_script(self, script_id, script):

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/scripts/"+script_id+"/content"

        body = {
            "script": script
        }
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_map(self):
        position = (len(self.components))
        map_id = self.add_component("map", position)
        return map_id

    def add_image(self, image, cta):
        position = (len(self.components))
        image_id = self.add_component("image", position)
        image_url = self.upload_image(image)
        self.edit_image(image_id, image_url, cta)
        return image_id


    def upload_image(self, image):

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata+'/smartconfiguration/icons?folder=insightdefinition'

        file = {'file': image}
        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=authentication_url, files=file, headers=headers)
        response.raise_for_status()
        return response.json()["url"]

    def edit_image(self, image_id, image_url, cta, image_name="Image 1"):

        body = {
            "id":image_id,
            "type":"image",
            "name":image_name,
            "customName":False,
            "dependsOn":[],
            "valid":True,
            "queryComponent":False,
            "link": cta,
            "image": image_url
        }

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/images/"+image_id
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

    def add_sql_query(self, query_sql, dataset_slug, is_native = False):

        position = (len(self.components))

        sql_id = self.add_component("sql_query", position)

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartdataset/datasets/slug/' + self.agent_id + '/' + dataset_slug
        #logging.info("AUTH URL {}".format(authentication_url))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=authentication_url, headers=headers)
        response.raise_for_status()
        r = response.json()
        dataset_id = r["dataset"]["id"]

        url_put = self.smart_insight_url + "/definitions/" + self.definition_id + "/sql_queries/" + sql_id
        body2 = {
            "datasetId": dataset_id,
            "sql": query_sql,
            "id": sql_id,
            "nativeType": is_native,
            "queryComponent": True,
            "type": "sql_query",
            "valid": True,
            "variableName": sql_id + "Result"
        }
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        r = s.put(url=url_put, json=body2, headers=headers)
        r.raise_for_status()

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/sql_queries/"+self.components[position]["id"]+"/sql"

        body = {
            "datasetId": dataset_id,
            "sql": query_sql,
        }

        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]

        has_result_set = True

        for component in r.json()["components"]:
            if component["id"] == sql_id:
                if component["preview"] == [] or component["preview"][0]["details"]["rows"] == []:
                    has_result_set = False


        return sql_id , has_result_set


    def delete(self):

        url= self.smart_insight_url+"/definitions/"+self.definition_id+"?final=true"
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }

        r = s.delete(url=url, headers=headers)
        r.raise_for_status()


    def add_query(self, dataset_slug, fields):

        '''
        fields = [{column: "STATUS_HISTORY_NUOVO_TIME", aggregation: null,…}}
        '''

        position = (len(self.components))

        # Get dataset_id
        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        authentication_url = self._base_url_askdata + '/smartdataset/datasets/slug/' + \
                             self.agent_id + '/' + dataset_slug

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=authentication_url, headers=headers)
        response.raise_for_status()
        r = response.json()

        try:
            dataset_id = r["dataset"]["id"]
        except KeyError:
            logging.error("DATASET NOT FOUND")
            return

        url_get = self.smart_insight_url + "/composed_queries?datasetId=" + dataset_id

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=url_get, headers=headers)
        response.raise_for_status()
        r = response.json()

        query_composer = r["qc"]

        url_preview = self.smart_insight_url+"/composed_queries/"+query_composer["id"]+"/preview?limit=100?&agentId="+self.agent_id

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=url_preview, json={}, headers=headers)
        response.raise_for_status()
        r = response.json()
        qc_fields = query_composer["fields"]

        new_fields = []

        for qc_field in qc_fields:
            for field in fields:
                if (field["column"] == qc_field["column"] or field["column"] == qc_field["alias"]):
                    qc_field["aggregation"] = field["aggregation"]
                    del qc_field["internalDataType"]
                    new_fields.append(qc_field)

        query_composer["fields"] = new_fields
        del query_composer["relationships"]
        query_composer["where"] = []
        query_composer["orderBy"] = []

        post_url = self.smart_insight_url+"/composed_queries"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=post_url, json=query_composer, headers=headers)
        response.raise_for_status()

        qc_post_url = self.smart_insight_url+"/definitions/"+self.definition_id+"/query_composers"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        body = {"type": "query_composer", "position": position, "qc": query_composer["id"]}

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.post(url=qc_post_url, json=body, headers=headers)
        response.raise_for_status()
        r = response.json()


    def add_search_query(self, query):

        position = (len(self.components))
        query_id = self.add_component("nl_query", position)

        body_query = {"nl": query, "language": "en"}

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        query_url = self.smart_insight_url +'/definitions/'+ self.definition_id +'/nl_queries/'+ query_id + '/nl'

        #logging.info("QUERY URL {}".format(query_url))
        r = s.put(url=query_url, json=body_query, headers=headers)

        self.components = r.json()["components"]

        return self.components[position]["id"]


    def add_component(self, type, position):

        body = {"type": type, "position": position}

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        query_url = self.smart_insight_url + '/definitions/' + self.definition_id + '/components/'
        #logging.info("URL {}".format(query_url))
        r = s.post(url=query_url, json=body, headers=headers)
        r.raise_for_status()
        self.components = r.json()["components"]
        return self.components[position]["id"]


    def get_url(self):
        url = self.app_url

        get_url = self._base_url_askdata+"/smartbot/agents/"+self.agent_id

        r = self.__make_get_request(get_url)

        agent_slug = r.json()["slug"]

        url += agent_slug

        url += "/card/"
        url += self.slug
        return url

    def add_admin(self, admin_email):

        get_url = self._base_url_askdata+"/smartbot/share/principals?q="+admin_email

        r = self.__make_get_request(get_url)

        if(r.json()!=[]):
            for user in r.json():
                body = {
                    "role": "ADMIN",
                    "userId": user["id"]
                }
                try:
                    url = self.smart_insight_url+"/insights/"+self.definition_id+"/users"

                    self.__make_post_request(url, body)

                except Exception:
                    logging.info("User {} can't be invited to channel, make sure the email is associated to an Askdata user".format(admin_email))

        else:
            logging.error("EMAIL ERRATA: Inserire una email corrispondente ad un utente Askdata")


    def add_follower(self, follower_email):

        get_url = self._base_url_askdata+"/smartbot/share/principals?q="+follower_email

        r = self.__make_get_request(get_url)
        if(r.json()!=[]):
            for user in r.json():
                body = {
                    "role": "FOLLOWER",
                    "userId": user["id"]
                }
                try:
                    url = self.smart_insight_url+"/insights/"+self.definition_id+"/users"

                    self.__make_post_request(url, body)
                except:
                    logging.info("User {} can't be invited to channel, make sure the email is associated to an Askdata user".format(follower_email))
        else:
            logging.error("EMAIL ERRATA: Inserire una email corrispondente ad un utente Askdata")


    def edit_card(self, icon="", name=""):

        if(icon==""):
            icon = self.icon
        else:
            self.icon=icon

        if(name==""):
            name = self.name
        else:
            self.name=name

        body={
            "icon": icon,
            "name": name,
            "slug": self.slug
        }

        url = self.smart_insight_url+"/definitions/"+self.definition_id
        r = self.__make_put_request(url, body)
        self.components = r.json()["components"]


    def get(self):

        get_url = self.smart_insight_url+"/definitions/"+self.definition_id+"/card"

        r = self.__make_get_request(get_url)

        return r.json()["attachment"]

    def delete_component(self, component_id):

        for component in self.components:
            if(component["id"]==component_id):
                self.components.remove(component)

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        query_url = self.smart_insight_url + '/definitions/' + self.definition_id + '/components/'+component_id
        logging.info("DELETE URL {}".format(query_url))
        r = s.delete(url=query_url, headers=headers)
        r.raise_for_status()

    def set_on_demand(self, enabled=True):

        body = {"enabled":enabled}

        url = self.smart_insight_url + "/definitions/" + self.definition_id+ "/intents/status"
        self.__make_put_request(url, body)

    def add_on_demand_queries(self, queries):

        body = {"userQuery": queries}

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/intents"
        self.__make_put_request(url, body)

    def schedule_daily(self, channel_slug, hours, minutues, seconds):

        cron = seconds + " " + minutues + " " + hours + " * * *"

        url = self._base_url_askdata + '/smartfeed/channels?agentId=' + self.agent_id + '&slug=' + channel_slug

        r = self.__make_get_request(url)

        channel_id = r.json()[0]["id"]

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/publishing/"

        body = {
            "schedulingCron": cron,
            "enabled": True,
            "destination": {"channels":[channel_id],"code":None,"name":None,"icon":None,"visibility":None,"queryId":None,"empty":None},
            "condition": "always",
            "advancedConditions": None}

        self.__make_post_request(url, body)


    def schedule_monthly(self, channel_slug, day_of_month, hours, minutues, seconds):

        """day_of_week: day of the week in format MON, TUE, WED, THU, FRI, SAT, SUN"""

        cron = seconds + " " + minutues + " " + hours + day_of_month + " * ?"

        url = self._base_url_askdata + '/smartfeed/channels?agentId=' + self.agent_id + '&slug=' + channel_slug
        r = self.__make_get_request(url)

        channel_id = r.json()[0]["id"]

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/publishing/"

        body = {
            "schedulingCron": cron,
            "enabled": True,
            "destination": {"channels":[channel_id],"code":None,"name":None,"icon":None,"visibility":None,"queryId":None,"empty":None},
            "condition": "always",
            "advancedConditions": None}

        self.__make_post_request(url, body)

    def schedule_weekly(self, channel_slug, day_of_week, hours, minutues, seconds):
        """day_of_week: day of the week in format MON, TUE, WED, THU, FRI, SAT, SUN"""

        cron = seconds + " " + minutues + " " + hours + " ? * " + day_of_week

        url = self._base_url_askdata + '/smartfeed/channels?agentId=' + self.agent_id + '&slug=' + channel_slug
        r = self.__make_get_request(url)

        channel_id = r.json()[0]["id"]

        url = self.smart_insight_url + "/definitions/" + self.definition_id + "/publishing/"

        body = {
            "schedulingCron": cron,
            "enabled": True,
            "destination": {"channels": [channel_id], "code": None, "name": None, "icon": None, "visibility": None,
                            "queryId": None, "empty": None},
            "condition": "always",
            "advancedConditions": None}

        self.__make_post_request(url, body)

    def clone(self, channel_id, card_name="", card_slug=None):

        if card_name == "":
            card_name = "Copy of " + self.name

        if card_slug is not None:
            try:
                url = self.smart_insight_url + "/definitions/agent/" + self.agent_id + "/slug/" + card_slug

                r = self.__make_get_request(url)

                definition = r.json()
                insight_definition = Insight_Definition(self._env, self._token, self.agent_slug, definition)
                insight_definition.delete()
            except Exception as e:
                logging.info("{} card not found... creating new one".format(card_slug))
                #logging.info("{}".format(e.__str__()))

        url = self.smart_insight_url+"/definitions/"+self.definition_id+"/clone"
        body = {
            "agentId": self.agent_id,
            "name": card_name,
            "slug": card_slug,
            "channelId": channel_id
            }

        self.__make_post_request(url, body)


import os
import sys
import json
import datetime
import requests
import pandas as pd
import numpy as np
import argparse
import logging
import logging.config

from googleapiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials

# Logging
logger = logging.getLogger(__name__)

# Constants
SCOPES = ['https://www.googleapis.com/auth/analytics.readonly']
KEY_FILE_LOCATION = './credentials/ga-api-key.json'
VIEW_ID = 'ga:157729614'

# Functions
def initialize_analyticsreporting(scopes, key_file_location):
    '''Initializes an Analytics Reporting API V4 service object.

    Returns:
      An authorized Analytics Reporting API V4 service object.
    '''
    credentials = ServiceAccountCredentials.from_json_keyfile_name(
        key_file_location, scopes)

    # Build the service object.
    analytics = build('analyticsreporting', 'v4', credentials=credentials)

    return analytics


''' Transform the response in a dataframe '''
def response_to_dataframe(response):
    ''' Parses and prints the Analytics Reporting API V4 response.

    Args:
      response: An Analytics Reporting API V4 response.
    '''
    list = []
    for report in response.get('reports', []):
        columnHeader = report.get('columnHeader', {})
        dimensionHeaders = columnHeader.get('dimensions', [])
        metricHeaders = columnHeader.get('metricHeader', {}).get('metricHeaderEntries', [])

        for row in report.get('data', {}).get('rows', []):
            dict = {}
            dimensions = row.get('dimensions', [])
            dateRangeValues = row.get('metrics', [])

            for header, dimension in zip(dimensionHeaders, dimensions):
                dict[header] = dimension

            for i, values in enumerate(dateRangeValues):
                for metric, value in zip(metricHeaders, values.get('values')):
                    #if ',' in value or ',' in value:
                    if ',' in value:
                        dict[metric.get('name')] = float(value)
                    else:
                        dict[metric.get('name')] = int(value)

            list.append(dict)

    df = pd.DataFrame(list)
    return df

def read_google_analytics(analytics, view_id, start_date, end_date):
    '''Queries the Analytics Reporting API V4.

    Args:
      analytics: An authorized Analytics Reporting API V4 service object.
    Returns:
      The Analytics Reporting API V4 response.
    '''
    response = analytics.reports().batchGet(
        body={
            'reportRequests': [
                {
                    'viewId': view_id,
                    'dateRanges': [{'startDate': start_date, 'endDate': end_date}],
                    'metrics': [{'expression': 'ga:sessions'},
                                {'expression': 'ga:users'},
                                {'expression': 'ga:avgSessionDuration'},
                                {'expression': 'ga:percentNewSessions'},
                                {'expression': 'ga:pageviews'},
                                {'expression': 'ga:uniquePageviews'},
                                {'expression': 'ga:avgTimeOnPage'},
                                {'expression': 'ga:bounceRate'},
                                {'expression': 'ga:exitRate'},
                                {'expression': 'ga:pageValue'},
                                {'expression': 'ga:transactions'},
                                {'expression': 'ga:transactionRevenue'},
                                {'expression': 'ga:transactionsPerSession'},
                                {'expression': 'ga:transactionsPerUser'},
                                {'expression': 'ga:goal1Completions'},
                                {'expression': 'ga:goal2Completions'},
                                {'expression': 'ga:goal3Completions'},
                                {'expression': 'ga:goal4Completions'},
                                {'expression': 'ga:goal5Completions'},
                                {'expression': 'ga:goal6Completions'},
                                {'expression': 'ga:goal7Completions'},
                                {'expression': 'ga:goal8Completions'},
                                {'expression': 'ga:goal9Completions'}],
                    'dimensions': [{'name': 'ga:date'},
                                   {'name': 'ga:sourceMedium'},
                                   {'name': 'ga:campaign'},
                                   {'name': 'ga:adContent'},
                                   {'name': 'ga:keyword'},
                                   {'name': 'ga:adGroup'},
                                   {'name': 'ga:deviceCategory'},
                                   {'name': 'ga:landingPagePath'},
                                   {'name': 'ga:exitPagePath'},
                                   {'name': 'ga:date'}]
                }]
        }
    ).execute()

    return response_to_dataframe(response)



import requests
import pandas as pd

''' Function that call the Tableau Server REST API to perform a query '''
def query_tableau(query, site_id, username, password):
    # Set the request parameters
    url = 'https://tableau.server.url/api/2.6/sites/' + site_id + '/workbooks?filter=' + query
    # Set proper headers
    headers = {'content-type': 'application/json', 'accept': 'application/json'}
    # Do the HTTP request
    response = requests.get(url, auth=(username, password), headers=headers)
    # Check for HTTP codes other than 200
    if response.status_code != 200:
        print('Status:', response.status_code, 'Headers:', response.headers, 'Error Response:', response.json())
        exit()
    # Decode the JSON response into a dictionary and use the data
    data = response.json()
    return data
    
''' Convert response in a dataframe '''
def get_dataframe(data):
    # Get the list of workbooks
    workbooks = data['workbooks']
    # Create a dataframe from the list
    df = pd.DataFrame(workbooks)
    return df




import os
import pandas as pd
import yaml
import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
import uuid
import getpass
import logging

root_dir = os.path.abspath(os.path.dirname(__file__))
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    url_list = yaml.load(file, Loader=yaml.FullLoader)

_LOG_FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s() ] - %(asctime)s --> %(message)s"
g_logger = logging.getLogger()
logging.basicConfig(format=_LOG_FORMAT)
g_logger.setLevel(logging.INFO)

@pd.api.extensions.register_dataframe_accessor("askdata")
class AskdataPandasExtension:
    def __init__(self, pandas_df):
        self._df:pd.DataFrame = pandas_df

        if(not hasattr(self, '_env')):
            self._env = input('Askdata Enviroment: ').lower()

        if self._env == 'dev':
            self.base_url_security = url_list['BASE_URL_SECURITY_DEV']

        if self._env == 'qa':
            self.base_url_security = url_list['BASE_URL_SECURITY_QA']

        if self._env == 'prod':
            self.base_url_security = url_list['BASE_URL_SECURITY_PROD']

        username = input('Askdata Username: ')

        password = getpass.getpass(prompt='Askdata Password: ')

        self.username = username

        data = {
            "grant_type": "password",
            "username": self.username,
            "password": password
        }

        headers = {
            "Authorization": "Basic YXNrZGF0YS1zZGs6YXNrZGF0YS1zZGs=",
            "Content-Type": "application/x-www-form-urlencoded",
            "cache-control": "no-cache,no-cache"
        }

        authentication_url = self.base_url_security + '/domain/askdata/oauth/token'

        # request token for the user
        s = requests.Session()
        s.keep_alive = False
        r1 = s.post(url=authentication_url, data=data, headers=headers)
        r1.raise_for_status()
        self._token = r1.json()['access_token']
        self.r1 = r1

    def normalize_columns(self):

        problematicChars = [",", ";", ":", "{", "}", "(", ")", "=", ">", "<", "."]
        new_cols = {}
        for column in self._df.columns:

            columnName = column.lower()

            for p_char in problematicChars:
                columnName = columnName.replace(p_char, "")

            columnName = columnName.replace(" ", "_")
            columnName = columnName.replace("-", "_")
            columnName = columnName.strip()

            for p_char in problematicChars:
                columnName = columnName.replace(p_char, "")

            new_cols[column] = columnName

        return self._df.rename(columns=new_cols, inplace=True)


    def save_dataset(self, workspace:str, dataset_slug="", replace=False, dataset_name="", locale="en"):

        self._df = self._df.where(pd.notnull(self._df), None)

        for col in self._df.columns:
            if (self._df[col].dtypes == "datetime64[ns]"):
                self._df[col] = self._df[col].astype("str")

        if self._env.lower() == 'dev':
            askdata_url = url_list['BASE_URL_ASKDATA_DEV']
        elif self._env.lower() == 'qa':
            askdata_url = url_list['BASE_URL_ASKDATA_QA']
        elif self._env.lower() == 'prod':
            askdata_url = url_list['BASE_URL_ASKDATA_PROD']

        if "/" in workspace:
            agentSlug = workspace.split("/")[0]
            datasetSlug = workspace.split("/")[1]
        else:
            agentSlug = workspace
            if dataset_slug == "":
                datasetSlug = "parquet_" + uuid.uuid4().__str__()
            else:
                datasetSlug = dataset_slug

        url = askdata_url + "/smartbot/agents/{}/datasets/{}/data_table".format(agentSlug, datasetSlug)

        fields= []
        for column in self._df.columns:
            fields.append({"name": column})

        schema = {"fields": fields}

        body = {
            "data": self._df.to_dict(orient="record"),
            "schema": schema,
            "locale": locale
        }

        if(replace):

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.put(url=url, json=body, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]

        else:
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.post(url=url, json=body, headers=headers)
            response.raise_for_status()
            dataset = response.json()
            dataset_id = dataset["id"]

        if(dataset_name!=""):
            url = askdata_url + "/smartbot/agents/"+dataset["agents"][0]+"/datasets/"+dataset_id+"/settings"
            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.get(url=url, headers=headers)
            response.raise_for_status()
            settings = response.json()

            settings["label"] = dataset_name

            s = requests.Session()
            s.keep_alive = False
            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
            s.mount('https://', HTTPAdapter(max_retries=retries))

            headers = {
                "Content-Type": "application/json",
                "Authorization": "Bearer" + " " + self._token
            }
            response = s.put(url=url,json=settings , headers=headers)
            response.raise_for_status()

        #Execute sync
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer" + " " + self._token
        }
        dataset_url = askdata_url + '/smartdataset/datasets/' + dataset_id + '/sync'
        r = requests.post(url=dataset_url, headers=headers)
        r.raise_for_status()


    def join(self, with_query:str, workspace:str, how="inner", on=None, left_on=None, right_on=None, lang="en"):

        if self._env.lower() == 'dev':
            askdata_url = url_list['BASE_URL_ASKDATA_DEV']
        elif self._env.lower() == 'qa':
            askdata_url = url_list['BASE_URL_ASKDATA_QA']
        elif self._env.lower() == 'prod':
            askdata_url = url_list['BASE_URL_ASKDATA_PROD']

        if(on==None and (left_on== None and right_on==None)):
            logging.error("Pleanse insert either 'on' condition or right_on and left_on conditions!")
            return

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = askdata_url + '/smartbot/agents/agentslug/' + workspace

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=authentication_url, headers=headers)
        response.raise_for_status()
        r = response.json()
        agent_id = r["agentId"]

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))
        authentication_url = askdata_url + '/security/me'

        headers = {"Authorization": "Bearer" + " " + self._token}
        response = s.get(url=authentication_url, headers=headers)
        response.raise_for_status()
        r = response.json()
        self.user_id = r["id"]

        url = askdata_url + "/smartbot/agents/" + agent_id + "/ask"

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        body = {
            "message": with_query,
            "language": lang.lower(),
            "userId": self.user_id
        }

        response = s.post(url=url, json=body, headers=headers)
        response.raise_for_status()
        result = response.json()["result"]
        df_query = pd.DataFrame(data=result[0]["attachment"]["body"][0]["details"]["rows"],
                                columns=result[0]["attachment"]["body"][0]["details"]["columns"])

        return self._df.merge(df_query, how, on, left_on, right_on)





import pandas as pd
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import logging
import requests
import os
import yaml

pd.set_option('display.max_columns', None)
pd.set_option('display.width', 200)
pd.set_option('display.max_rows', 10)

root_dir = os.path.abspath(os.path.dirname(__file__))
yaml_path = os.path.join(root_dir, '../askdata/askdata_config/base_url.yaml')
with open(yaml_path, 'r') as file:
    url_list = yaml.load(file, Loader=yaml.FullLoader)


# SmartJoin
def smart_join(tables):

    if bool(tables):

        headers = {
            "Content-Type": "application/json"
        }

        data = {
            "tables": tables
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = url_list['BASE_URL_SMARTJOIN_DEV'] + "/get_joins"

        r = s.post(url=url, headers=headers, json=data)
        r.raise_for_status()

        try:
            response = r.json()['joins']
            return response
        except Exception as e:
            logging.error(str(e))
            print(e)
            return None
    else:
        print("The input is empty!")
        return None


# SmartSynonyms
def smart_synonyms(word, lang=None, initial_synonyms_list=None):

    if lang is None:
        lang = "en"
        print("Using English language as default!")

    if initial_synonyms_list is None:
        initial_synonyms_list = []

    # Google Pod
    headers = {
        "Content-Type": "application/json"
    }

    s = requests.Session()
    s.keep_alive = False
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    s.mount('https://', HTTPAdapter(max_retries=retries))

    url = url_list['BASE_URL_SMART_SYNONYM_PROD']

    data = {
        "word": word,
        "lang": lang,
        "initial_synonyms_list ": initial_synonyms_list
    }

    r = s.post(url=url, headers=headers, json=data)
    r.raise_for_status()

    try:
        dict_response = r.json()
        synonyms = dict_response['synonyms']
        business_name = dict_response['business_name']
        return synonyms, business_name
    except Exception as e:
        logging.error(str(e))


# SmartOpenData
def smart_opendata(sentence, language="en", show_header=False, dataset_score_dict=None, boost_columns=False,
                boost_name_score=False, use_domains=True, store_df=True):
    if language != "en" and language != "it":
        print("Please use 'en' for english or 'it' for italian.")
        dict_response = {}
        return dict_response

    if sentence != "":
        headers = {
            "Content-Type": "application/json"
        }

        data = {
            "nl": sentence,
            "lang": language,
            "dataset_score_dict": dataset_score_dict,
            "store_df": store_df,
            "use_domains": use_domains,
            "boost_columns": boost_columns,
            "boost_name_score": boost_name_score
        }

        s = requests.Session()
        s.keep_alive = False
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
        s.mount('https://', HTTPAdapter(max_retries=retries))

        url = url_list['BASE_URL_OPENDATA_DEV'] + "/spacy"

        r = s.post(url=url, headers=headers, json=data)
        r.raise_for_status()

        try:
            dict_response = r.json()

            if show_header and store_df:
                # Print header
                i = 1
                j = 0
                for el in dict_response['dataset_rank']["first_place"]["dataset_list"]:
                    score = dict_response['dataset_rank']["first_place"]["scoring"]
                    name = el['dataset']
                    df = pd.DataFrame().from_dict(dict_response['df_list'][j])
                    print("#------------------------------------------------------------------#")
                    print("Dataset " + name + " ranked as #" + str(i) + " with score: " + str(score) + ".")
                    print()
                    print(df.head())
                    print("#------------------------------------------------------------------#")
                    print()
                    j += 1
                # i += 1

            return dict_response
        except Exception as e:
            logging.error(str(e))
    else:
        print("Input sentence is empty!")
        dict_response = {}
        return dict_response


def opendata_specific_result(response, request, show=True):
    possible_request = ["first_place", "second_place", "third_place", "top_places", "all_datasets", "header", "joins",
                        "top1_df"]
    explained_possible_request = ["'first_place' - to obtain info of rank#1 datasets",
                                  "'second_place' - to obtain info of rank#2 datasets",
                                  "'third_place' - to obtain info of rank#3 datasets",
                                  "'top_places' - to obtain info of top ranked#1-3 datasets",
                                  "'all_datasets' - to obtain info of all ranked datasets",
                                  "'header' - to print rank#1 datasets headers",
                                  "'joins' - to obtain the possible joins between all ranked datasets",
                                  "'top1_df' - to obtain the top-1 dataset as a DataFrame"]

    if request in possible_request:

        # Print first three places
        if "first_place" in request or "second_place" in request or "third_place" in request:

            if request == "first_place":
                dataset_infos = response["dataset_rank"]['first_place']
                i = 1
            elif request == "second_place":
                dataset_infos = response["dataset_rank"]['second_place']
                i = 2
            else:
                dataset_infos = response["dataset_rank"]['third_place']
                i = 3

            if show:
                print("#----------------------------------------------------------------------------------------#")
                print("Rank#" + str(i) + " datasets")
                print()
                # Print first place datasets name
                for el in dataset_infos["dataset_list"]:
                    print("#-----------------------------#")
                    print("Dataset info:")
                    print()
                    print("Dataset name: " + el['dataset'])  # name
                    print()
                    print("Detected columns: " + str(el['important_columns']))  # detected columns
                    print()
                    print("All columns: " + str(el["columns"]))  # all columns
                    print()
                    # Print place scoring
                    print("Score: " + str(dataset_infos["scoring"]))
                    print("#-----------------------------#")
                    print()
                print("#----------------------------------------------------------------------------------------#")

            return dataset_infos

        # Print all top places
        elif request == "top_places":
            if show:
                for i in range(1, 4):

                    print("#----------------------------------------------------------------------------------------#")
                    print("Rank#" + str(i) + " datasets: ")
                    print()

                    if i == 1:
                        dataset_infos = response["dataset_rank"]["first_place"]
                    elif i == 2:
                        dataset_infos = response["dataset_rank"]["second_place"]
                    else:
                        dataset_infos = response["dataset_rank"]["third_place"]

                    # Print first place datasets name
                    for el in dataset_infos["dataset_list"]:
                        print("#-----------------------------#")
                        print("Dataset info:")
                        print()
                        print("Dataset name: " + el['dataset'])  # name
                        print()
                        print("Detected columns: " + str(el['important_columns']))  # detected columns
                        print()
                        print("All columns: " + str(el["columns"]))  # all columns
                        print()
                        # Print place scoring
                        print("Score: " + str(dataset_infos["scoring"]))
                        print("#-----------------------------#")
                        print()
                print("#----------------------------------------------------------------------------------------#")
            return response["dataset_rank"]

        elif request == "header":
            try:
                i = 1
                j = 0
                for el in response['dataset_rank']["first_place"]["dataset_list"]:
                    score = response['dataset_rank']["first_place"]["scoring"]
                    name = el['dataset']
                    df = pd.DataFrame().from_dict(response['df_list'][j])
                    if show:
                        print("#----------------------------------------------------------------------------------------#")
                        print("Dataset " + name + " ranked as #" + str(i) + " with score: " + str(score) + ".")
                        print()
                        print(df.head())
                        print("#----------------------------------------------------------------------------------------#")
                        print()
                    j += 1
                    return df.head()
            except:
                print("There is no DataFrame in the input structure. Use 'store_df' when you get the structure.")
                return None

        elif request == "all_datasets":

            # Print first, second third and other datasets places names
            all_datasets = response["dataset_list"]

            if show:
                print("#----------------------------------------------------------------------------------------#")
                print()
                print("List of all main datasets detected:")
                print()

                for el in all_datasets:
                    print(el['dataset'])  # name

                print("#----------------------------------------------------------------------------------------#")
                print()

            return all_datasets

        elif request == "joins":
            # Print the possible joins between datasets detected
            joins = response["join_dict"]

            if show:
                print("#----------------------------------------------------------------------------------------#")
                print()
                print("List of all joins between datasets detected:")
                print()
                for column, datasets in joins.items():
                    print("Column: " + column)
                    print("Datasets: " + str(datasets))
                    print()
                print("#----------------------------------------------------------------------------------------#")
                print()
            return joins

        elif request == "top1_df":
            try:
                df = pd.DataFrame().from_dict(response['df_list'][0])
                return df
            except:
                print("There is no DataFrame in the input structure. Use 'store_df' when you get the structure.")
                return pd.DataFrame()
    else:
        print("Please use a request from the following:")
        for el in explained_possible_request:
            print(el)
        return None



from askdata.smartgraph import smart_join

if __name__ == "__main__":
    input = "left:[(cinema id, numeric, ['1', '1', '1', '2', '6']);(film id, numeric, ['1', '2', '3', '1', '5']);(date, string, ['21 may', '21 may', '21 jun', '11 july', '2 aug']);(show times per day, numeric, ['5', '3', '2', '5', '4']);(price, numeric, ['12.99', '12.99', '8.99', '9.99', '12.99'])]; right:[(film id, numeric, ['1', '2', '3', '4', '5']);(rank in series, numeric, ['26', '27', '28', '29', '30']);(number in season, numeric, ['1', '2', '3', '4', '5']);(title, string, ['the case of the mystery weekend', 'the case of the smart dummy', 'the case: off the record', 'the case of the bermuda triangle', 'the case of the piggy banker']);(directed by, string, ['bill schreiner', 'bill schreiner', 'bill schreiner', 'jesus salvador treviño', 'bill schreiner']);(original air date, string, ['september 21–25, 1992', 'september 28–october 2, 1992', 'october 5–9, 1992', 'october 12–16, 1992', 'october 19–23, 1992']);(production code, string, ['50021–50025', '50231–50235', '50011–50015', '50251–50255', '50241–50245']) ]"
    response = smart_join(input)
    print(response)



##required !pip install facebook_business -q

from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.business import Business
from facebook_business.adobjects.adaccountuser import AdAccountUser
from facebook_business.adobjects.adsinsights import AdsInsights
from facebook_business.adobjects.adreportrun import AdReportRun
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.adset import AdSet


import datetime
import time
import sys
import pandas as pd
import _locale



#return pandas dataframe

#See here for documentation https://developers.facebook.com/docs/marketing-api/insights/parameters/ 



"""


MANDATORY INPUT: my_app_id, my_app_secret, my_access_token, start_date, end_date,


"""



def get_fb_ads(my_app_id, my_app_secret, my_access_token, start_date, end_date, account_name):
    
    
    _locale._getdefaultlocale = (lambda *args: ['en_US', 'UTF-8'])


    # Start the connection to the facebook API
    FacebookAdsApi.init(my_app_id, my_app_secret, my_access_token, api_version='v13.0')

    # Create a business object for the business account
    business = Business('WAF')


    # Get today's date for the filename, and the csv data
    yesterdaybad= datetime.datetime.now() 
    yesterdayslash = yesterdaybad.strftime('%m/%d/%Y')
    yesterdayhyphen = yesterdaybad.strftime('%m-%d-%Y')

    # Get all ad accounts on the business account
    me = AdAccountUser(fbid='me')
    accounts = list(me.get_ad_accounts())

    params = {
        "time_range" : {"since": start_date, "until": end_date},
        'fields': [
                    AdsInsights.Field.account_id,
                    AdsInsights.Field.account_name,
                    AdsInsights.Field.action_values,
                    AdsInsights.Field.actions,
                    AdsInsights.Field.ad_id,
                    AdsInsights.Field.ad_name,
                    AdsInsights.Field.adset_id,
                    AdsInsights.Field.adset_name,
                    AdsInsights.Field.campaign_id,
                    AdsInsights.Field.campaign_name, 
                    AdsInsights.Field.cost_per_action_type,
                    AdsInsights.Field.cost_per_inline_link_click,
                    AdsInsights.Field.cost_per_inline_post_engagement,
                    AdsInsights.Field.cost_per_unique_click,
                    AdsInsights.Field.cpm, 
                    AdsInsights.Field.cpp, 
                    AdsInsights.Field.ctr, 
                    AdsInsights.Field.date_start,
                    AdsInsights.Field.date_stop,
                    AdsInsights.Field.frequency,
                    AdsInsights.Field.impressions,
                    AdsInsights.Field.inline_link_clicks,
                    AdsInsights.Field.inline_post_engagement,
                    AdsInsights.Field.reach,
                    AdsInsights.Field.spend,
                    AdsInsights.Field.unique_clicks,
                    AdsInsights.Field.unique_ctr,
                    AdsInsights.Field.video_p100_watched_actions,
                    AdsInsights.Field.video_p25_watched_actions,
                    AdsInsights.Field.video_p50_watched_actions,
                    AdsInsights.Field.video_p75_watched_actions,
                    AdsInsights.Field.video_p95_watched_actions,
                    AdsInsights.Field.website_ctr,
                    AdsInsights.Field.clicks
          ],
        'level': 'campaign',
        'time_increment': 1
    }


    records = []

    #Wait till all the data is pulled 
    def wait_for_async_job(async_job):
        async_job.api_get()
        while async_job[AdReportRun.Field.async_status] != 'Job Completed' or async_job[AdReportRun.Field.async_percent_completion] < 100:
            time.sleep(5)
            async_job.api_get()
    r=0    
    


    # Iterate through the accounts
    records = []
    for k, v in account_name.items() :
        print("Account: '{}', Id: '{}'".format(k, v))
        tempaccount = AdAccount(v)
        try:
            ads = tempaccount.get_insights_async(fields = params.get('fields'), params=params) 
            wait_for_async_job(ads)
            x = ads.get_result()
        except:
            pass
        for ad in x:
            # Set default values in case the insight info is empty
            date_of_report = yesterdaybad
            account_id = ""
            account_name = ""
            # action_values = ""
            fb_pixel_lead = ""
            lead = ""
            actions = ""
            actions_link_click = ""
            actions_post_reaction = ""
            actions_post = ""
            actions_onsite_conversion_post_save = ""
            actions_page_engagement = ""
            actions_post_engagement = ""
            ad_id = ""
            ad_name = ""
            adset_id = ""
            adset_name = ""
            campaign_id = ""
            campaign_name = "" 
            cost_per_action_type = ""
            cost_per_action_type_link_click = ""
            cost_per_action_type_post_reaction = ""
            cost_per_action_type_post = ""
            cost_per_action_type_onsite_conversion_post_save = ""
            cost_per_action_type_page_engagement = ""
            cost_per_action_type_post_engagement = ""
            cost_per_inline_link_click = ""
            cost_per_inline_post_engagement = ""
            cost_per_unique_click = ""
            cpm = ""
            cpp = ""
            ctr = ""
            date_start = ""
            date_stop = ""
            frequency = ""
            impressions = ""
            inline_link_clicks = ""
            inline_post_engagement = ""
            reach = ""
            spend = ""
            unique_clicks = ""
            unique_ctr = ""
            video_p100_watched_actions = ""
            video_p25_watched_actions = ""
            video_p50_watched_actions = ""
            video_p75_watched_actions = ""
            video_p95_watched_actions = ""
            website_ctr = ""
            website_ctr_link_click = ""
            clicks = ""

            # Set values from insight data
            if ('account_id' in ad) :
                account_id = ad['account_id']
            if ('account_name' in ad) :
                account_name = ad['account_name']
            if ('action_values' in ad) :
                # action_values= ad['action_values']
                for i in ad['action_values']:
                    if i['action_type'] == 'offsite_conversion.fb_pixel_lead':
                        fb_pixel_lead = i['value']
                    if i['action_type'] == 'lead':
                        lead = i['value']
            if ('actions' in ad) :
                for i in ad['actions']:
                    if i['action_type'] == 'link_click':
                        actions_link_click = i['value']
                    if i['action_type'] == 'post_reaction':
                        actions_post_reaction = i['value']
                    if i['action_type'] == 'post':
                        actions_post = i['value']
                    if i['action_type'] == 'onsite_conversion.post_save':
                        actions_onsite_conversion_post_save = i['value']
                    if i['action_type'] == 'page_engagement':
                        actions_page_engagement = i['value']
                    if i['action_type'] == 'post_engagement':
                        actions_post_engagement = i['value']            
            if ('ad_id' in ad) :
                ad_id= ad['ad_id']
            if ('ad_name' in ad) :
                ad_name= ad['ad_name']
            if ('adset_id' in ad) :
                adset_id= ad['adset_id']
            if ('adset_name' in ad) :
                adset_name= ad['adset_name']
            if ('campaign_id' in ad) :
                campaign_id= ad['campaign_id']
            if ('campaign_name' in ad) :
                campaign_name= ad['campaign_name'] 
            if ('cost_per_action_type' in ad) :
                for i in ad['cost_per_action_type']:
                    if i['action_type'] == 'link_click':
                        cost_per_action_type_link_click = i['value']
                    if i['action_type'] == 'post_reaction':
                        cost_per_action_type_post_reaction = i['value']
                    if i['action_type'] == 'post':
                        cost_per_action_type_post = i['value']
                    if i['action_type'] == 'onsite_conversion.post_save':
                        cost_per_action_type_onsite_conversion_post_save = i['value']
                    if i['action_type'] == 'page_engagement':
                        cost_per_action_type_page_engagement = i['value']
                    if i['action_type'] == 'post_engagement':
                        cost_per_action_type_post_engagement = i['value']    
            if ('cost_per_inline_link_click' in ad) :
                cost_per_inline_link_click= ad['cost_per_inline_link_click']
            if ('cost_per_inline_post_engagement' in ad) :
                cost_per_inline_post_engagement= ad['cost_per_inline_post_engagement']
            if ('cost_per_unique_click' in ad) :
                cost_per_unique_click= ad['cost_per_unique_click']
            if ('cpm' in ad) :
                cpm= ad['cpm']
            if ('cpp' in ad) :
                cpp= ad['cpp']
            if ('ctr' in ad) :
                ctr= ad['ctr']
            if ('date_start' in ad) :
                date_start= ad['date_start']
            if ('date_stop' in ad) :
                date_stop= ad['date_stop']
            if ('frequency' in ad) :
                frequency= ad['frequency']
            if ('impressions' in ad) :
                impressions= ad['impressions']
            if ('inline_link_clicks' in ad) :
                inline_link_clicks= ad['inline_link_clicks']
            if ('inline_post_engagement' in ad) :
                inline_post_engagement= ad['inline_post_engagement']
            if ('reach' in ad) :
                reach= ad['reach']
            if ('spend' in ad) :
                spend= ad['spend']
            if ('unique_clicks' in ad) :
                unique_clicks= ad['unique_clicks']
            if ('unique_ctr' in ad) :
                unique_ctr= ad['unique_ctr']
            if ('video_p100_watched_actions' in ad) :
                for i in ad['video_p100_watched_actions']:
                    if i['action_type'] == 'video_view':
                        video_p100_watched_actions = i['value']                
            if ('video_p25_watched_actions' in ad) :
                for i in ad['video_p25_watched_actions']:
                    if i['action_type'] == 'video_view':
                        video_p25_watched_actions = i['value']             

            if ('video_p50_watched_actions' in ad) :
                for i in ad['video_p50_watched_actions']:
                    if i['action_type'] == 'video_view':
                        video_p50_watched_actions = i['value']             

            if ('video_p75_watched_actions' in ad) :
                for i in ad['video_p75_watched_actions']:
                    if i['action_type'] == 'video_view':
                        video_p75_watched_actions = i['value']   

            if ('video_p95_watched_actions' in ad) :
                for i in ad['video_p95_watched_actions']:
                    if i['action_type'] == 'video_view':
                        video_p95_watched_actions = i['value'] 

            if ('website_ctr' in ad) :
                for i in ad['website_ctr']:
                    if i['action_type'] == 'link_click':
                        website_ctr_link_click = i['value']    
            if ('date_of_report' in ad) :
                date_of_report = ad['date_of_report']
            if ('clicks' in ad) :
                clicks = ad['clicks']
            #if r%15 == 0:
             #   print(r)
            r = r+1
            # Write all ad info to the file, and increment the number of rows that will display
            records.append({
                "account_id" : account_id,
                "account_name" : account_name,
                "fb_pixel_lead" : fb_pixel_lead,
                "lead" : lead,
                "actions_link_click" : actions_link_click ,
                "actions_post_reaction " : actions_post_reaction ,
                "actions_post" : actions_post,
                "actions_onsite_conversion_post_save" : actions_onsite_conversion_post_save,
                "actions_page_engagement" : actions_page_engagement,
                "actions_post_engagement" : actions_post_engagement,
                "ad_id" : ad_id,
                "ad_name" : ad_name,
                "adset_id" : adset_id,
                "adset_name" : adset_name,
                "campaign_id" : campaign_id,
                "campaign_name" : campaign_name,
                "cost_per_action_type_link_click" : cost_per_action_type_link_click ,
                "cost_per_action_type_post_reaction" : cost_per_action_type_post_reaction,
                "cost_per_action_type_post" : cost_per_action_type_post,
                "cost_per_action_type_onsite_conversion_post_save" : cost_per_action_type_onsite_conversion_post_save,
                "cost_per_action_type_page_engagement" : cost_per_action_type_page_engagement,
                "cost_per_action_type_post_engagement" : cost_per_action_type_post_engagement,
                "cost_per_inline_link_click" : cost_per_inline_link_click,
                "cost_per_inline_post_engagement" : cost_per_inline_post_engagement,
                "cost_per_unique_click" : cost_per_unique_click,
                "cpm" : cpm,
                "cpp" : cpp,
                "ctr" : ctr,
                "date_start" : date_start,
                "date_stop" : date_stop,
                "frequency" : frequency,
                "impressions" : impressions,
                "inline_link_clicks" : inline_link_clicks,
                "inline_post_engagement" : inline_post_engagement,
                "reach" : reach,
                "spend" : spend,
                "unique_clicks" : unique_clicks,
                "unique_ctr" : unique_ctr,
                "video_p100_watched_actions" : video_p100_watched_actions,
                "video_p25_watched_actions" : video_p25_watched_actions,
                "video_p50_watched_actions" : video_p50_watched_actions,
                "video_p75_watched_actions" : video_p75_watched_actions,
                "video_p95_watched_actions" : video_p95_watched_actions,
                "website_ctr_link_click" : website_ctr_link_click,
                "clicks" : clicks,
                "date_of_report" : date_of_report
                    })

    fb_ads = pd.DataFrame(records)

    return fb_ads
